#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
GraphQL 查询语句文件差异报告生成模块

此模块提供了差异报告生成功能，主要包括：
1. Markdown格式的差异报告生成
2. HTML格式的差异报告生成

这些功能用于在GraphQL Schema发生变更时，生成详细的差异比较报告，
便于用户直观地了解API变更情况，实现API变更的可视化跟踪。
"""

import datetime
import difflib


def generate_markdown_diff_report(diff_result, old_content, new_content, old_file, new_file):
    """
    基于结构化的差异数据生成Markdown格式报告
    
    Args:
        diff_result (dict): 差异结果字典
        old_content (dict): 旧文件内容
        new_content (dict): 新文件内容
        old_file (str): 旧文件路径
        new_file (str): 新文件路径
        
    Returns:
        str: Markdown格式的差异报告
    """    
    # 生成详细的文本差异
    diff_lines = []
    diff_lines.append("# GraphQL API 变更监测报告")
    diff_lines.append(f"## 旧文件: {old_file}")
    diff_lines.append(f"## 新文件: {new_file}")
    diff_lines.append("")
    
    # 添加项目级差异信息
    if diff_result["added_projects"]:
        diff_lines.append("## 新增项目")
        for project in diff_result["added_projects"]:
            diff_lines.append(f"- {project}")
        diff_lines.append("")
    
    if diff_result["removed_projects"]:
        diff_lines.append("## 移除项目")
        for project in diff_result["removed_projects"]:
            diff_lines.append(f"- {project}")
        diff_lines.append("")
    
    # 添加查询语句差异信息
    if diff_result["modified_projects"]:
        diff_lines.append("## 修改项目")
        for project, changes in diff_result["modified_projects"].items():
            diff_lines.append(f"### {project}")
            
            if changes["added_queries"]:
                diff_lines.append("#### 新增查询")
                for query in changes["added_queries"]:
                    diff_lines.append(f"- {query}")
                    diff_lines.append(f"  ```")
                    diff_lines.append(f"  {new_content[project][query]}")
                    diff_lines.append(f"  ```")
                diff_lines.append("")
            
            if changes["removed_queries"]:
                diff_lines.append("#### 移除查询")
                for query in changes["removed_queries"]:
                    diff_lines.append(f"- {query}")
                    diff_lines.append(f"  ```")
                    diff_lines.append(f"  {old_content[project][query]}")
                    diff_lines.append(f"  ```")
                diff_lines.append("")
            
            if changes["modified_queries"]:
                diff_lines.append("#### 修改查询")
                for query in changes["modified_queries"]:
                    diff_lines.append(f"- {query}")
                    diff_lines.append(f"  1. 旧语句")
                    diff_lines.append(f"  ```")
                    diff_lines.append(f"  {old_content[project][query]}")
                    diff_lines.append(f"  ```")
                    diff_lines.append(f"  2. 新语句")
                    diff_lines.append(f"  ```")
                    diff_lines.append(f"  {new_content[project][query]}")
                    diff_lines.append(f"  ```")
                    
                    # 生成详细的文本差异
                    old_lines = old_content[project][query].splitlines()
                    new_lines = new_content[project][query].splitlines()
                    differ = difflib.Differ()
                    diff = list(differ.compare(old_lines, new_lines))
                    if len(diff) > 1:  # 如果有多于一行的差异
                        diff_lines.append(f"  3. 差异详情")
                        diff_lines.append(f"  ```diff")
                        for line in diff:
                            diff_lines.append(f"  {line}")
                        diff_lines.append(f"  ```")
                diff_lines.append("")
            
            diff_lines.append("")
    
    return diff_lines

def generate_html_diff_report(diff_result, old_content, new_content, old_file, new_file):
    """
    基于结构化的差异数据生成HTML格式报告
    
    Args:
        diff_result (dict): 差异结果字典
        old_content (dict): 旧文件内容
        new_content (dict): 新文件内容
        old_file (str): 旧文件路径
        new_file (str): 新文件路径
        
    Returns:
        str: HTML格式的差异报告
    """
    # 获取当前时间，用于报告标题
    current_time = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
    
    # HTML头部
    html = [
        "<!DOCTYPE html>",
        "<html lang='zh-CN'>",
        "<head>",
        "    <meta charset='utf-8'>",
        "    <meta name='viewport' content='width=device-width, initial-scale=1.0'>",
        "    <title>GraphQL API 变更监测报告</title>",
        "    <!-- Tailwind CSS CDN -->",
        "    <script src='https://cdn.tailwindcss.com'></script>",
        "    <!-- FontAwesome CDN -->",
        "    <link rel='stylesheet' href='https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.4.0/css/all.min.css'>",
        "    <!-- 代码高亮 Highlight.js -->",
        "    <link rel='stylesheet' href='https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.7.0/styles/github.min.css'>",
        "    <script src='https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.7.0/highlight.min.js'></script>",
        "    <script src='https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.7.0/languages/graphql.min.js'></script>",
        "    <style>",
        "        /* 自定义样式补充 */",
        "        .diff-added { color: #22863a; background-color: #e6ffed; padding: 1px 5px; white-space: pre; }",
        "        .diff-removed { color: #cb2431; background-color: #ffeef0; padding: 1px 5px; white-space: pre; }",
        "        .diff-unchanged { color: #6a737d; padding: 1px 5px; white-space: pre; }",
        "        .code-block { font-family: ui-monospace, SFMono-Regular, Menlo, Monaco, Consolas, monospace; position: relative; }",
        "        .code-block > div { line-height: 1.5; white-space: pre; }",
        "        .collapsible-content { max-height: 0; overflow: hidden; transition: max-height 0.3s ease-out; }",
        "        .expanded .collapsible-content { max-height: 2000px; transition: max-height 0.5s ease-in; }",
        "        .code-container { position: relative; }",
        "        .copy-btn { position: absolute; top: 5px; right: 5px; background: rgba(255,255,255,0.8); border: none; border-radius: 4px; padding: 3px 8px; cursor: pointer; font-size: 12px; display: flex; align-items: center; box-shadow: 0 1px 3px rgba(0,0,0,0.1); }",
        "        .copy-btn:hover { background: rgba(255,255,255,1); }",
        "        .copy-btn i { margin-right: 3px; }",
        "    </style>",
        "    <script>",
        "        // 页面加载完成后初始化",
        "        document.addEventListener('DOMContentLoaded', function() {",
        "            // 初始化代码高亮",
        "            document.querySelectorAll('pre code').forEach(block => {",
        "                hljs.highlightElement(block);",
        "            });",
        "            ",
        "            // 初始化折叠面板",
        "            document.querySelectorAll('.collapsible-header').forEach(header => {",
        "                header.addEventListener('click', function() {",
        "                    this.parentElement.classList.toggle('expanded');",
        "                    const icon = this.querySelector('.toggle-icon');",
        "                    if (icon) {",
        "                        icon.classList.toggle('fa-chevron-down');",
        "                        icon.classList.toggle('fa-chevron-right');",
        "                    }",
        "                });",
        "            });",
        "            ",
        "            // 添加复制按钮到所有代码块",
        "            document.querySelectorAll('.code-container').forEach(container => {",
        "                const copyBtn = container.querySelector('.copy-btn');",
        "                if (copyBtn) {",
        "                    copyBtn.addEventListener('click', function() {",
        "                        const codeElement = container.querySelector('pre code') || container.querySelector('.code-block');",
        "                        let text = '';",
        "                        ",
        "                        if (codeElement.tagName.toLowerCase() === 'code') {",
        "                            text = codeElement.innerText;",
        "                        } else {",
        "                            // 对于差异详情，我们只获取非注释行的文本",
        "                            const divs = codeElement.querySelectorAll('div:not(.diff-unchanged)');",
        "                            const lines = Array.from(divs).map(div => {",
        "                                // 移除差异标记",
        "                                return div.innerText;",
        "                            });",
        "                            text = lines.join('\\n');",
        "                        }",
        "                        ",
        "                        navigator.clipboard.writeText(text).then(() => {",
        "                            const icon = this.querySelector('i');",
        "                            const originalClass = icon.className;",
        "                            icon.className = 'fas fa-check';",
        "                            this.classList.add('bg-green-100');",
        "                            setTimeout(() => {",
        "                                icon.className = originalClass;",
        "                                this.classList.remove('bg-green-100');",
        "                            }, 2000);",
        "                        }, () => {",
        "                            alert('复制失败，请手动复制');",
        "                        });",
        "                    });",
        "                }",
        "            });",
        "        });",
        "    </script>",
        "</head>",
        "<body class='bg-gray-50 text-gray-900 min-h-screen'>",
        "    <div class='container mx-auto px-4 py-8 max-w-6xl'>",
        "        <!-- 报告标题 -->",
        "        <header class='mb-8'>",
        "            <div class='flex items-center justify-between'>",
        "                <h1 class='text-3xl font-bold text-indigo-700'>",
        "                    <i class='fas fa-code-compare mr-2'></i>GraphQL API 变更监测报告",
        "                </h1>",
        "                <span class='text-gray-500'><i class='far fa-clock mr-1'></i>" + current_time + "</span>",
        "            </div>",
        "            <div class='mt-4 p-4 bg-white rounded-lg shadow-sm border-l-4 border-indigo-500'>",
        "                <div class='grid grid-cols-1 md:grid-cols-2 gap-4'>",
        "                    <div>",
        "                        <p class='text-sm text-gray-500'>旧文件:</p>",
        "                        <p class='font-mono text-sm truncate'>" + old_file + "</p>",
        "                    </div>",
        "                    <div>",
        "                        <p class='text-sm text-gray-500'>新文件:</p>",
        "                        <p class='font-mono text-sm truncate'>" + new_file + "</p>",
        "                    </div>",
        "                </div>",
        "            </div>",
        "        </header>",
        "",
        "        <!-- 报告内容 -->",
        "        <main class='space-y-6'>"
    ]
    
    # 处理新增项目
    if diff_result["added_projects"]:
        html.append("""
            <section class='bg-white rounded-lg shadow-sm overflow-hidden'>
                <h2 class='text-xl font-semibold p-4 bg-gray-100 flex items-center'>
                    <i class='fas fa-square-plus text-green-600 mr-2'></i>新增项目
                </h2>
                <div class='p-4 space-y-2'>
        """)
        
        for project in diff_result["added_projects"]:
            html.append(f"""
                    <div class='p-2 bg-gray-50 rounded-md flex items-center'>
                        <i class='fas fa-folder-plus text-green-500 mr-2'></i>
                        <span>{project}</span>
                    </div>
            """)
        
        html.append("</div></section>")
    
    # 处理移除项目
    if diff_result["removed_projects"]:
        html.append("""
            <section class='bg-white rounded-lg shadow-sm overflow-hidden'>
                <h2 class='text-xl font-semibold p-4 bg-gray-100 flex items-center'>
                    <i class='fas fa-square-minus text-red-600 mr-2'></i>移除项目
                </h2>
                <div class='p-4 space-y-2'>
        """)
        
        for project in diff_result["removed_projects"]:
            html.append(f"""
                    <div class='p-2 bg-gray-50 rounded-md flex items-center'>
                        <i class='fas fa-folder-minus text-red-500 mr-2'></i>
                        <span>{project}</span>
                    </div>
            """)
        
        html.append("</div></section>")
    
    # 处理修改项目
    if diff_result["modified_projects"]:
        html.append("""
            <section class='bg-white rounded-lg shadow-sm overflow-hidden'>
                <h2 class='text-xl font-semibold p-4 bg-gray-100 flex items-center'>
                    <i class='fas fa-pen-to-square text-blue-600 mr-2'></i>修改项目
                </h2>
                <div class='p-4'>
        """)
        
        for project, changes in diff_result["modified_projects"].items():
            html.append(f"""
                <section class='bg-white rounded-lg shadow-sm overflow-hidden mt-4'>
                    <h3 class='text-lg font-medium p-3 bg-blue-50 border-l-4 border-blue-500 flex items-center justify-between'>
                        <div><i class='fas fa-project-diagram mr-2'></i>{project}</div>
                        <span class='text-sm text-gray-500'>项目</span>
                    </h3>
                    <div class='p-4'>
            """)
            
            # 处理新增查询
            if changes["added_queries"]:
                html.append("""
                        <div class='mb-4'>
                            <h4 class='text-md font-medium mb-2 flex items-center text-green-600'>
                                <i class='fas fa-plus-circle mr-2'></i>新增查询
                            </h4>
                            <div class='space-y-3'>
                """)
                
                for query in changes["added_queries"]:
                    query_content = new_content[project][query]
                    escaped_content = query_content.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;")
                    
                    html.append(f"""
                                <div class='border border-gray-200 rounded-md overflow-hidden'>
                                    <div class='collapsible-header cursor-pointer bg-gray-50 p-2 flex justify-between items-center hover:bg-gray-100'>
                                        <span class='font-medium'>{query}</span>
                                        <i class='fas fa-chevron-down toggle-icon text-gray-500'></i>
                                    </div>
                                    <div class='collapsible-content'>
                                        <div class='p-3'>
                                            <div class='code-container'>
                                                <button class='copy-btn'><i class='far fa-copy'></i>复制</button>
                                                <pre class='m-0 p-0 bg-gray-50'><code class='language-graphql p-3 block overflow-x-auto'>{escaped_content}</code></pre>
                                            </div>
                                        </div>
                                    </div>
                                </div>
                    """)
                
                html.append("</div></div>")
            
            # 处理移除查询
            if changes["removed_queries"]:
                html.append("""
                        <div class='mb-4'>
                            <h4 class='text-md font-medium mb-2 flex items-center text-red-600'>
                                <i class='fas fa-minus-circle mr-2'></i>移除查询
                            </h4>
                            <div class='space-y-3'>
                """)
                
                for query in changes["removed_queries"]:
                    query_content = old_content[project][query]
                    escaped_content = query_content.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;")
                    
                    html.append(f"""
                                <div class='border border-gray-200 rounded-md overflow-hidden'>
                                    <div class='collapsible-header cursor-pointer bg-gray-50 p-2 flex justify-between items-center hover:bg-gray-100'>
                                        <span class='font-medium'>{query}</span>
                                        <i class='fas fa-chevron-down toggle-icon text-gray-500'></i>
                                    </div>
                                    <div class='collapsible-content'>
                                        <div class='p-3'>
                                            <div class='code-container'>
                                                <button class='copy-btn'><i class='far fa-copy'></i>复制</button>
                                                <pre class='m-0 p-0 bg-gray-50'><code class='language-graphql p-3 block overflow-x-auto'>{escaped_content}</code></pre>
                                            </div>
                                        </div>
                                    </div>
                                </div>
                    """)
                
                html.append("</div></div>")
            
            # 处理修改查询
            if changes["modified_queries"]:
                html.append("""
                        <div class='mb-4'>
                            <h4 class='text-md font-medium mb-2 flex items-center text-amber-600'>
                                <i class='fas fa-pencil mr-2'></i>修改查询
                            </h4>
                            <div class='space-y-3'>
                """)
                
                for query in changes["modified_queries"]:
                    old_query_content = old_content[project][query]
                    new_query_content = new_content[project][query]
                    
                    # 转义内容
                    old_escaped = old_query_content.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;")
                    new_escaped = new_query_content.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;")
                    
                    # 生成差异内容
                    old_lines = old_query_content.splitlines()
                    new_lines = new_query_content.splitlines()
                    differ = difflib.Differ()
                    diff = list(differ.compare(old_lines, new_lines))
                    
                    # 差异内容HTML
                    diff_html = []
                    for line in diff:
                        if line.startswith('+ '):
                            diff_html.append(f"<div class='diff-added'>{line[2:].replace('&', '&amp;').replace('<', '&lt;').replace('>', '&gt;')}</div>")
                        elif line.startswith('- '):
                            diff_html.append(f"<div class='diff-removed'>{line[2:].replace('&', '&amp;').replace('<', '&lt;').replace('>', '&gt;')}</div>")
                        elif line.startswith('? '):
                            diff_html.append(f"<div class='diff-unchanged'>{line.replace('&', '&amp;').replace('<', '&lt;').replace('>', '&gt;')}</div>")
                        else:
                            diff_html.append(f"<div>{line[2:].replace('&', '&amp;').replace('<', '&lt;').replace('>', '&gt;')}</div>")
                    
                    html.append(f"""
                                <div class='border border-gray-200 rounded-md overflow-hidden'>
                                    <div class='collapsible-header cursor-pointer bg-gray-50 p-2 flex justify-between items-center hover:bg-gray-100'>
                                        <span class='font-medium'>{query}</span>
                                        <i class='fas fa-chevron-down toggle-icon text-gray-500'></i>
                                    </div>
                                    <div class='collapsible-content'>
                                        <div class='p-3'>
                                            <div class='mt-2 mb-1 flex items-center'>
                                                <i class='fas fa-history text-blue-600 mr-2'></i>
                                                <span class='font-medium'>1. 旧语句</span>
                                            </div>
                                            <div class='code-container'>
                                                <button class='copy-btn'><i class='far fa-copy'></i>复制</button>
                                                <pre class='m-0 p-0 bg-gray-50'><code class='language-graphql p-3 block overflow-x-auto'>{old_escaped}</code></pre>
                                            </div>
                                            
                                            <div class='mt-2 mb-1 flex items-center'>
                                                <i class='fas fa-code text-blue-600 mr-2'></i>
                                                <span class='font-medium'>2. 新语句</span>
                                            </div>
                                            <div class='code-container'>
                                                <button class='copy-btn'><i class='far fa-copy'></i>复制</button>
                                                <pre class='m-0 p-0 bg-gray-50'><code class='language-graphql p-3 block overflow-x-auto'>{new_escaped}</code></pre>
                                            </div>
                    """)
                    
                    if len(diff) > 1:
                        html.append(f"""
                                            <div class='mt-2 mb-1 flex items-center'>
                                                <i class='fas fa-code-compare text-purple-600 mr-2'></i>
                                                <span class='font-medium'>3. 差异详情</span>
                                            </div>
                                            <div class='code-container'>
                                                <button class='copy-btn'><i class='far fa-copy'></i>复制</button>
                                                <div class='code-block p-3 bg-gray-50 overflow-x-auto whitespace-pre' style="max-width: 100%;">{''.join(diff_html)}</div>
                                            </div>
                        """)
                    
                    html.append("""
                                        </div>
                                    </div>
                                </div>
                    """)
                
                html.append("</div></div>")
            
            html.append("</div></section>")
        
        html.append("</div></section>")
    
    # 添加页脚和关闭标签
    html.extend([
        "        </main>",
        "",
        "        <!-- 页脚 -->",
        "        <footer class='mt-8 pt-4 border-t border-gray-200 text-center text-gray-500 text-sm'>",
        "            <p>由 GraphQL Schema to HttpRunner 工具生成</p>",
        f"            <p class='mt-1'>© {datetime.datetime.now().year} GraphQL to HttpRunner</p>",
        "        </footer>",
        "    </div>",
        "</body>",
        "</html>"
    ])
    
    return "\n".join(html)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
GraphQL Schema to HttpRunner 测试用例转换工具

该模块是GraphQL Schema到HttpRunner测试用例转换工具的主入口。
处理命令行参数，读取GraphQL Schema文件，协调调用解析器和生成器模块。

主要功能：
1. 解析命令行参数，提供友好的命令行界面
2. 读取GraphQL Schema文件或通过内省查询获取Schema
3. 协调调用SchemaParser解析Schema
4. 协调调用TestCaseGenerator生成测试用例
5. 协调调用QueryGenerator生成查询语句列表
6. 支持通过配置文件批量生成多个项目的测试用例
7. 支持指定单个查询名称生成特定查询或测试用例
"""

import argparse
import sys
import os
import csv
import time
import shutil

from . import __version__
from .parser import GraphQLSchemaParser
from .generator import HttpRunnerTestCaseGenerator
from .query_generator import GraphQLQueryGenerator
from .introspection import fetch_schema_from_introspection, IntrospectionQueryError
from .utils import backup_queries_file, compare_query_files, auto_update_testcases


def process_single_project(introspection_url=None, schema_file=None, output='query.yml', base_url="http://localhost:8888",
                         max_depth=2, is_api=False, is_cite=False, required=False, is_skip=False, is_testcases=True, project_name=None, query_name=None):
    """处理单个项目"""
    schema = None
    
    # 生成用例时，重置默认输出目录
    if is_testcases and output == 'query.yml':
        if is_api:
            output = 'api'
        else:
            output = 'testcases'
    
    # 从Schema文件中读取
    if schema_file:
        # 检查Schema文件是否存在
        if not os.path.isfile(schema_file):
            print(f"错误：Schema文件 '{schema_file}' 不存在")
            return False
        
        # 读取Schema文件
        try:
            with open(schema_file, 'r', encoding='utf-8') as f:
                schema_content = f.read()
        except Exception as e:
            print(f"读取Schema文件时出错: {e}")
            return False
        
        # 解析Schema
        print(f"开始解析GraphQL Schema文件: {schema_file}")
        try:
            parser = GraphQLSchemaParser(schema_content)
            schema = parser.parse()
        except Exception as e:
            print(f"解析Schema文件时出错: {e}")
            return False
    
    # 通过内省查询获取Schema
    elif introspection_url:
        try:
            schema = fetch_schema_from_introspection(introspection_url)
        except IntrospectionQueryError as e:
            print(f"内省查询失败: {e}")
            return False
        except Exception as e:
            print(f"获取Schema时出错: {e}")
            return False
    
    # 如果指定了查询名称，检查它是否存在于Schema中
    if query_name and query_name not in schema.root_fields:
        print(f"错误：指定的查询名称 '{query_name}' 在Schema中不存在")
        print(f"可用的查询名称有: {', '.join(schema.root_fields.keys())}")
        return False
    
    # 生成测试用例
    if is_testcases:
        output_type = "API层" if is_api else "用例层"
        print(f"\n开始生成HttpRunner {output_type}测试用例...")
        try:
            generator = HttpRunnerTestCaseGenerator(schema, base_url, max_depth, required, is_skip, is_cite)

            if is_api:
                if query_name:
                    # 只生成指定查询的API层测试用例
                    testcase_count = generator.generate_single_api_test_case(output, query_name)
                    print(f"\n已生成{testcase_count}个API层测试用例到目录: {output}")
                else:
                    testcase_count = generator.generate_api_test_cases(output)
                    print(f"\n已生成{testcase_count}个API层测试用例到目录: {output}")
            else:
                if query_name:
                    # 只生成指定查询的用例层测试用例
                    testcase_count = generator.generate_single_test_case(output, query_name)
                    print(f"\n已生成{testcase_count}个用例层测试用例到目录: {output}")
                else:
                    testcase_count = generator.generate_test_cases(output)
                    print(f"\n已生成{testcase_count}个用例层测试用例到目录: {output}")

        except Exception as e:
            print(f"生成测试用例时出错: {e}")
            return False
    
    # 生成查询语句列表
    else:
        print(f"\n开始生成GraphQL查询语句列表...")
        try:
            generator = GraphQLQueryGenerator(schema, max_depth)
            
            if query_name:
                # 只生成指定查询的查询语句
                queries = generator.generate_queries(output, project_name, query_name)
                print(f"\n已生成{query_name}查询语句到文件: {output}")
            else:
                queries = generator.generate_queries(output, project_name)
                query_count = len(queries)
                print(f"\n已生成{query_count}个查询语句到文件: {output}")
        except Exception as e:
            print(f"生成查询语句时出错: {e}")
            return False
    
    print(f"使用的最大查询深度: {max_depth}")
    print(f"使用的基础URL或产品名: {base_url}")
    if is_testcases:
        print(f"是否只包含必选参数：{'是' if required else '否'}")
        print(f"是否包含skip关键词：{'是' if is_skip else '否'}")
        if is_api:
            print(f"是否生成引用API层测试用例：{'是' if is_cite else '否'}")
            if is_cite:
                print(f"引用API层测试用例输出目录：{output.replace('api', 'testcases', 1) if 'api' in output else output}")
    
    return True


def batch_generate(config_file, is_testcases=True, queries_file='query.yaml', report_format=None, is_auto_update=False):
    """
    批量生成HttpRunner测试用例
    
    Args:
        config_file (str): 批量处理配置文件路径
        is_testcases (bool): 是否生成测试用例
        queries_file (str): 指定批量处理模式生成查询语句文件路径，默认为query.yaml，当查询语句文件已存在时，会进行备份、与新文件对比并生成差异报告，以及可以更新测试用例
        report_format (str): 生成查询语句文件差异报告格式，可选值为'markdown'、'html'、'both'
        is_auto_update (bool): 是否自动更新测试用例
    Returns:
        True or dict: 批量生成成功返回True，当批量生成查询语句列表且相对旧文件有变更时，返回差异变更结果字典，用于更新测试用例
    """
    
    if not os.path.exists(config_file):
        print(f"错误：配置文件 '{config_file}' 不存在")
        sys.exit(1)
    
    print(f"开始批量生成测试用例，配置文件：{config_file}")
    
    # 读取配置文件
    projects = []
    try:
        with open(config_file, "r", encoding="utf-8") as f:
            reader = csv.DictReader(f)
            for row in reader:
                projects.append(row)
    except Exception as e:
        print(f"读取配置文件时出错: {e}")
        sys.exit(1)
    
    if not projects:
        print("配置文件中没有发现项目")
        sys.exit(1)
    
    backup_file = None
    if not is_testcases and os.path.exists(queries_file):
        # 备份旧查询语句文件
        backup_file = backup_queries_file(queries_file)
    
    # 批量处理每个项目
    success_count = 0
    failed_count = 0
    skipped_count = 0
    for project in projects:
        project_name = project["project_name"]
        
        # 检查项目是否已下线，如果已下线则跳过
        if project.get("offline", "false").lower() == "true":
            print(f"\n{'='*80}")
            print(f"跳过已下线项目: {project_name}")
            skipped_count += 1
            continue
            
        introspection_url = project["introspection_url"]
        output = project["output"] if is_testcases else queries_file
        base_url = project["base_url"]
        max_depth = int(project.get("max_depth", 2))
        required = project.get("is_required", "false").lower() == "true"
        is_api = project.get("is_api", "false").lower() == "true"
        is_cite = project.get("is_cite", "false").lower() == "true"
        is_skip = project.get("is_skip", "false").lower() == "true"
        
        print(f"\n{'='*80}")
        print(f"开始处理项目: {project_name}")
        
        # 处理单个项目
        start_time = time.time()
        result = process_single_project(
            introspection_url=introspection_url,
            output=output,
            base_url=base_url,
            max_depth=max_depth,
            is_api=is_api,
            is_cite=is_cite,
            required=required,
            is_skip=is_skip,
            is_testcases=is_testcases,
            project_name=project_name
        )
        
        if result:
            success_count += 1
            end_time = time.time()
            print(f"生成完成，耗时: {end_time - start_time:.2f}秒")
        else:
            failed_count += 1
            print(f"生成失败，请稍后重新生成！")
        
    # 输出总结
    print(f"\n{'='*80}")
    print(f"批量生成任务完成")
    print(f"成功处理项目数: {success_count}")
    print(f"失败处理项目数: {failed_count}")
    print(f"跳过已下线项目数: {skipped_count}")
    print(f"总项目数: {len(projects)}")

    # 如果存在失败项目，则退出程序
    if failed_count > 0:
        # 如果生成查询语句文件失败且有备份文件，则删除新文件并恢复备份文件
        if not is_testcases and backup_file:
            if os.path.exists(queries_file):
                os.remove(queries_file)
            if backup_file:
                shutil.copy2(backup_file, queries_file)
                os.remove(backup_file)
            print(f"注意：生成查询语句文件时，存在失败项目，已删除新文件并恢复备份文件！")
        sys.exit(1)

    # 生成查询语句文件差异报告和返回结果差异信息
    if report_format:
        print(f"\n{'='*80}")
        return compare_query_files(backup_file, queries_file, report_format)

    # 自动更新用例时不生成差异报告，只返回差异结果信息场景
    if is_auto_update:
        print(f"\n{'='*80}")
        return compare_query_files(backup_file, queries_file, report_format)

    return True


def main():
    """主函数"""
    parser = argparse.ArgumentParser(description='将GraphQL Schema转换为HttpRunner测试用例或查询语句')
    
    # 添加版本信息选项
    parser.add_argument('-V', '--version', action='store_true', help='显示版本信息')
    
    # 添加批处理配置文件选项
    parser.add_argument('-b', '--batch', help='批量处理配置文件路径，如 config.csv')
    
    # 创建互斥组，schema文件和内省查询URL只能二选一
    source_group = parser.add_mutually_exclusive_group()
    source_group.add_argument('-f', '--schema-file', help='GraphQL Schema文件路径')
    source_group.add_argument('-i', '--introspection-url', help='GraphQL内省查询URL，如http://localhost:9527/graphql')
    
    # 创建互斥组，生成测试用例或查询语句列表
    output_type_group = parser.add_mutually_exclusive_group()
    output_type_group.add_argument('-t', '--testcases', action='store_true', help='生成HttpRunner测试用例')
    output_type_group.add_argument('-q', '--queries', action='store_true', help='生成GraphQL查询语句列表')
    
    parser.add_argument('-o', '--output', default='query.yml', help='生成结果输出路径，根据生成类型默认为api、testcases、query.yml三种情况') # 批量处理模式生成的查询语句文件默认为query.yaml
    parser.add_argument('-u', '--base-url', default='http://localhost:8888', help='API基础URL或项目名，项目名用来作为自定义函数参数生成API基础URL')
    parser.add_argument('-d', '--max-depth', type=int, default=2, help='GraphQL查询嵌套的最大深度，默认为2')
    parser.add_argument('--api', action='store_true', help='生成API层测试用例而非用例层测试用例')
    parser.add_argument('--cite', action='store_true', help='生成引用API层测试用例（与--api选项一起使用）')
    parser.add_argument('--required', action='store_true', help='只包含必选参数，默认情况下包含所有参数')
    parser.add_argument('--skip', action='store_true', help='生成的测试用例是否包含skip关键词')
    parser.add_argument('--project', default='project_name', help='指定项目名称')
    parser.add_argument('--report', choices=['markdown', 'html', 'both'], help='生成差异报告格式：markdown/html/both（与-q选项或-a选项一起使用）')
    parser.add_argument('-a', '--auto-update', action='store_true', help='自动更新测试用例（与-b选项一起使用）')
    parser.add_argument('--queries-file', default='query.yaml', help='指定批模式时查询语句文件生成路径，默认为query.yaml (与-b选项一起使用)')
    parser.add_argument('--query-name', help='指定要生成的单个查询名称，只生成该查询的测试用例或查询语句')
    
    args = parser.parse_args()
    
    # 如果指定了版本信息选项，显示版本信息后退出
    if args.version:
        print(f"{__version__}")
        return
    
    if args.report:
        if args.testcases:
            print("参数错误：生成查询语句文件差异报告需要与-q选项或-a选项一起使用!")
            sys.exit(1)
        if args.batch:
            if not os.path.exists(args.queries_file):
                print(f"参数错误：查询语句文件 '{args.queries_file}' 不存在!")
                sys.exit(1)
        else:
            if not os.path.exists(args.output):
                print(f"参数错误：查询语句文件 '{args.output}' 不存在!")
                sys.exit(1)

    # 如果指定了自动更新选项
    if args.auto_update:
        if not args.batch:
            print("参数错误：自动更新功能需要与-b/--batch选项一起使用")
            sys.exit(1)
        if not os.path.exists(args.queries_file):
            print(f"参数错误：查询语句文件 '{args.queries_file}' 不存在!")
            sys.exit(1)
        
        diff_result = batch_generate(args.batch, is_testcases=False, queries_file=args.queries_file, report_format=args.report, is_auto_update=True)
        if isinstance(diff_result, dict):
            auto_update_testcases(diff_result, args.batch, args.queries_file) # 有差异，根据变更内容自动更新测试用例
        else:
            if diff_result:
                print("无差异，无需更新测试用例!")
            else:
                print("查询语句差异对比出错，请排查后重试!")
                sys.exit(1)
        return # 不执行后续逻辑
    
    # 如果指定了批处理配置文件，进入批处理模式
    if args.batch:
        is_testcases = not args.queries  # 默认生成测试用例，除非指定了 -q 选项
        batch_result = batch_generate(args.batch, is_testcases, args.queries_file, args.report)
        if not isinstance(batch_result, dict):
            sys.exit(1) # 无差异结果信息返回退出码为1，方便Jenkins命令行集成
        return # 不执行后续逻辑
    
    # 检查常规模式下必需的参数
    if not args.schema_file and not args.introspection_url:
        parser.error("参数错误：必须指定 -f/--schema-file 或 -i/--introspection-url 选项，或者使用 -b/--batch 批处理配置文件")
        
    if not args.testcases and not args.queries:
        parser.error("参数错误：必须指定 -t/--testcases 或 -q/--queries 选项")
    
    backup_file = None
    # 单项目模式下，如果生成查询语句并且输出文件已存在，则备份旧文件
    if args.queries and os.path.exists(args.output):
        backup_file = backup_queries_file(args.output)
    
    # 处理单个项目
    result = process_single_project(
        introspection_url=args.introspection_url,
        schema_file=args.schema_file,
        output=args.output,
        base_url=args.base_url,
        max_depth=args.max_depth,
        is_api=args.api,
        is_cite=args.cite,
        required=args.required,
        is_skip=args.skip,
        is_testcases=args.testcases,
        project_name=args.project,
        query_name=args.query_name
    )
    
    # 如果生成失败，则退出程序
    if not result:
        # 如果生成查询语句文件失败且有备份文件，则删除新文件并恢复备份文件
        if args.queries and backup_file:
            if os.path.exists(args.output):
                os.remove(args.output)
            if backup_file:
                shutil.copy2(backup_file, args.output)
                os.remove(backup_file)
            print(f"注意：生成查询语句文件失败，已删除新文件并恢复备份文件！")
        sys.exit(1)

    # 生成查询语句文件差异报告
    if args.report:
        single_result = compare_query_files(backup_file, args.output, args.report)
        if not isinstance(single_result, dict):
            sys.exit(1) # 无差异结果信息返回退出码为1，方便Jenkins命令行集成


if __name__ == '__main__':
    main() 


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
HttpRunner 测试用例生成器

该模块提供了从GraphQL Schema生成HttpRunner测试用例的功能。它根据GraphQL Schema中
定义的查询、变更和订阅操作，生成对应的HttpRunner YAML格式测试用例文件。

主要功能：
1. 为GraphQL操作生成HttpRunner测试用例
2. 支持生成查询、变更和订阅操作的测试用例
3. 根据GraphQL类型自动生成合适的示例值
4. 递归处理嵌套类型，生成适当的查询字段
5. 生成符合HttpRunner格式要求的YAML测试用例
6. 支持生成指定单个查询的测试用例
"""

import os
import yaml
import random
import datetime
from typing import Dict, List, Any, Set
import re

from .models import GraphQLSchema


class HttpRunnerTestCaseGenerator:
    """HttpRunner测试用例生成器"""
    
    def __init__(self, schema: GraphQLSchema, base_url: str = "http://localhost:8888", max_depth: int = 2, required_only: bool = False, is_skip: bool = False, is_cite: bool = False):
        self.schema = schema
        self.base_url = base_url
        self.max_depth = max_depth
        self.required_only = required_only       # 是否只包含必选参数，默认为False（包含所有参数）
        self.is_skip = is_skip                   # 是否包含skip关键词，默认为False
        self.is_cite = is_cite                   # 是否生成API层引用测试用例，默认为False
        self.scalar_types = schema.scalar_types  # 使用从schema中解析到的标量类型集合
        self.enum_types = schema.enum_types      # 添加对枚举类型的引用
        self.generated_types: Set[str] = set()
        
    def generate_test_cases(self, output_dir: str):
        """生成所有查询操作的用例层测试用例"""
        return self._generate_cases(output_dir, is_api_layer=False)
    
    def generate_api_test_cases(self, output_dir: str):
        """生成所有查询操作的API层测试用例"""
        return self._generate_cases(output_dir, is_api_layer=True)
    
    def generate_single_test_case(self, output_dir: str, query_name: str):
        """生成单个查询操作的用例层测试用例
        
        Args:
            output_dir: 输出目录路径
            query_name: 要生成的查询名称
            
        Returns:
            int: 生成的测试用例数量，成功为1，失败为0
        """
        return self._generate_single_case(output_dir, query_name, is_api_layer=False)
    
    def generate_single_api_test_case(self, output_dir: str, query_name: str):
        """生成单个查询操作的API层测试用例
        
        Args:
            output_dir: 输出目录路径
            query_name: 要生成的查询名称
            
        Returns:
            int: 生成的测试用例数量，成功为1，失败为0
        """
        return self._generate_single_case(output_dir, query_name, is_api_layer=True)
    
    def _generate_single_case(self, output_dir: str, query_name: str, is_api_layer: bool = False):
        """生成单个查询测试用例的通用方法
        
        Args:
            output_dir: 输出目录路径
            query_name: 要生成的查询名称
            is_api_layer: 是否生成API层测试用例，默认为False（生成用例层测试用例）
            
        Returns:
            int: 生成的测试用例数量，成功为1，失败为0
        """
        os.makedirs(output_dir, exist_ok=True)
        
        # 检查查询是否存在
        if query_name not in self.schema.root_fields:
            print(f"错误：查询名称 '{query_name}' 在Schema中不存在")
            return 0
        
        field_info = self.schema.root_fields[query_name]
        # 保存测试用例
        self._save_test_case(output_dir, is_api_layer, query_name, field_info)
        print(f"生成测试用例: {output_dir}")
        return 1
    
    def _generate_cases(self, output_dir: str, is_api_layer: bool = False):
        """生成测试用例的通用方法
        
        Args:
            output_dir: 输出目录路径
            is_api_layer: 是否生成API层测试用例，默认为False（生成用例层测试用例）
            
        Returns:
            int: 生成的测试用例数量
        """
        os.makedirs(output_dir, exist_ok=True)
        
        # 记录生成的测试用例数量
        testcase_count = 0
        
        # 生成根查询测试用例
        for field_name, field_info in self.schema.root_fields.items():
            # 跳过名为 _ 的字段，这些字段通常只是占位符，没有实际测试意义
            if field_name == '_':
                print(f"跳过生成下划线字段测试用例: {field_name}")
                continue

            # 保存测试用例
            self._save_test_case(output_dir, is_api_layer, field_name, field_info)
            # 增加计数器
            testcase_count += 1
        
        # 返回生成的测试用例数量
        return testcase_count
    
    def _save_test_case(self, output_dir: str, is_api_layer: bool, field_name: str, field_info: Dict[str, Any]):
        """保存测试用例"""
        # 确定操作类型前缀
        operation_type = self._determine_operation_type(field_name)

        # 根据层级生成相应的测试用例
        if is_api_layer:
            api_definition, api_cite = self._generate_api_definition(field_name, field_info, output_dir)
            with open(os.path.join(output_dir, f"{operation_type}_{field_name}_api.yml"), 'w', encoding='utf-8') as f:
                yaml.dump(api_definition, f, default_flow_style=False, sort_keys=False, allow_unicode=True, width=9999)
            if api_cite:
                cite_output_dir = output_dir.replace("api", "testcases", 1) if 'api' in output_dir else output_dir
                os.makedirs(cite_output_dir, exist_ok=True)
                with open(os.path.join(cite_output_dir, f"{operation_type}_{field_name}_test.yml"), 'w', encoding='utf-8') as f:
                    yaml.dump(api_cite, f, default_flow_style=False, sort_keys=False, allow_unicode=True, width=9999)
        else:
            test_case = self._generate_query_test_case(field_name, field_info)
            with open(os.path.join(output_dir, f"{operation_type}_{field_name}_test.yml"), 'w', encoding='utf-8') as f:
                yaml.dump(test_case, f, default_flow_style=False, sort_keys=False, allow_unicode=True, width=9999)


    def _determine_operation_type(self, field_name: str) -> str:
        """确定字段所属的操作类型（查询、变更或订阅）"""
            
        # 检查字段是否定义在Mutation类型中
        if self.schema.mutation_type and field_name in getattr(self.schema.types.get(self.schema.mutation_type), 'fields', {}):
            return "mutation"
        # 检查字段是否定义在Subscription类型中
        elif self.schema.subscription_type and field_name in getattr(self.schema.types.get(self.schema.subscription_type), 'fields', {}):
            return "subscription"
        # 默认为查询操作
        else:
            return "query"
    
    def _compress_query(self, query: str) -> str:
        """
        压缩查询语句，移除多余空白，使其一行显示
        
        Args:
            query: 原始查询语句
            
        Returns:
            str: 压缩后的查询语句
        """
        # 替换所有连续空白为单个空格
        query = re.sub(r'\s+', ' ', query)
        # 确保花括号前后有空格
        query = re.sub(r'{\s*', '{ ', query)
        query = re.sub(r'\s*}', ' }', query)
        return query.strip()

    def _prepare_graphql_query(self, field_name: str, field_info: Dict[str, Any]) -> tuple:
        """准备GraphQL查询信息
        
        Args:
            field_name: 字段名称
            field_info: 字段信息
            
        Returns:
            tuple: 包含以下元素的元组
                - operation_type: 操作类型（查询、变更或订阅）
                - operation_name: 操作名称（"query"、"mutation"或"subscription"）
                - operation_description: 操作描述（中文）
                - variables: 变量字典
                - query: 构造的GraphQL查询语句
        """
        # 确定操作类型
        operation_type = self._determine_operation_type(field_name)
        operation_name = "query" if operation_type == "query" else "mutation" if operation_type == "mutation" else "subscription"
        operation_description = "查询" if operation_type == "query" else "变更" if operation_type == "mutation" else "订阅"
        
        # 构建查询参数
        variables = {}
        args_str = ""
        var_defs_parts = []  # 用于保存变量定义部分
        
        if field_info['args']:
            args_parts = []
            for arg_name, arg_info in field_info['args'].items():
                # 获取原始参数类型
                arg_type = arg_info['type']
                
                # 默认为所有参数生成适合该类型的示例值
                variables[arg_name] = self._generate_example_value(arg_type)

                # 当只需要为必选参数生成示例值，则删除可选参数示例值
                if self.required_only and not arg_type.endswith('!'):
                    del variables[arg_name]
                
                # 构建参数字符串
                args_parts.append(f"{arg_name}: ${arg_name}")
                
                # 添加变量定义，使用原始类型
                var_defs_parts.append(f"${arg_name}: {arg_type}")

            if args_parts:
                args_str = f"({', '.join(args_parts)})"
        
        # 构建变量定义部分
        var_defs_str = f"({', '.join(var_defs_parts)})" if var_defs_parts else ""
        
        # 构建查询字段
        return_type = field_info['type']
        fields_str = self._generate_fields_for_type(return_type)
        
        # 构建GraphQL查询
        query = f"""{operation_name} {field_name}{var_defs_str} {{
  {field_name}{args_str} {{
{fields_str}
  }}
}}"""
        
        # 压缩查询为一行
        query = self._compress_query(query)
        
        # 转义GraphQL查询中的$符号，避免与HttpRunner变量语法冲突
        query = query.replace("$", "$$")
        
        return operation_type, operation_name, operation_description, variables, query

    def _generate_query_test_case(self, field_name: str, field_info: Dict[str, Any]) -> Dict[str, Any]:
        """生成查询测试用例"""
        # 准备GraphQL查询
        _, _, operation_description, variables, query = self._prepare_graphql_query(field_name, field_info)
        
        # 构建HttpRunner测试用例
        test_case = {
            'config': {
                'name': f"{field_name} {operation_description}测试",
                'base_url': self.base_url if self.base_url.startswith('http') else f"${{get_config({self.base_url},graphql_url)}}",
                'variables': {
                    'user': 'xxx_user' if self.base_url.startswith('http') else f"${{get_config({self.base_url},v0_user)}}",
                    'pwd': 'xxx_pwd' if self.base_url.startswith('http') else f"${{get_config({self.base_url},v0_pwd)}}",
                    'sessionId': 'xxx' if self.base_url.startswith('http') else f"${{get_login({self.base_url},$user,$pwd,{self.base_url}_token)}}"
                }
            },
            'teststeps': [
                {
                    'name': f"执行 {field_name} {operation_description}",
                    'request': {
                        'method': 'POST',
                        'url': "/graphql",
                        'headers': {
                            'Content-Type': 'application/json; charset=utf-8',
                            'Accept-Language': 'zh',
                            'x-operation-name': f"{field_name}"
                        },
                        'cookies': {
                            'sessionId': '$sessionId'
                        },
                        'json': {
                            'operationName': f"{field_name}",
                            'query': query,
                            'variables': variables
                        }
                    },
                    'extract': [{'data': 'content.data'}],
                    'validate': [
                        {'eq': ['status_code', 200]},
                        {'contains': ['headers.Content-Type', 'application/json']},
                        {'ne': ['content.data', None]}
                    ]
                }
            ]
        }
        if self.is_skip:
            api_items = list(test_case['teststeps'][0].items())
            api_items.insert(1, ('skipIf', '${skip_test_in_production_env()}'))
            test_case['teststeps'][0] = dict(api_items)

        return test_case

    def _generate_api_definition(self, field_name: str, field_info: Dict[str, Any], output_dir: str) -> tuple:
        """生成API层定义"""
        # 准备GraphQL查询
        operation_type, _, operation_description, variables, query = self._prepare_graphql_query(field_name, field_info)

        # 构建API定义
        api_definition = {
            'name': f"{field_name}_{operation_description}",
            'base_url': self.base_url if self.base_url.startswith('http') else f"${{get_config({self.base_url},graphql_url)}}",
            'variables': {
                'user': 'xxx_user' if self.base_url.startswith('http') else f"${{get_config({self.base_url},v0_user)}}",
                'pwd': 'xxx_pwd' if self.base_url.startswith('http') else f"${{get_config({self.base_url},v0_pwd)}}",
                'sessionId': 'xxx' if self.base_url.startswith('http') else f"${{get_login({self.base_url},$user,$pwd,{self.base_url}_token)}}"
            },
            'request': {
                'method': 'POST',
                'url': "/graphql",
                'headers': {
                    'Content-Type': 'application/json; charset=utf-8',
                    'Accept-Language': 'zh',
                    'x-operation-name': f"{field_name}"
                },
                'cookies': {
                    'sessionId': '$sessionId'
                },
                'json': {
                    'operationName': f"{field_name}",
                    'query': query,
                    'variables': variables
                }
            },
            'validate': [
                {'eq': ['status_code', 200]},
                {'contains': ['headers.Content-Type', 'application/json']},
                {'ne': ['content.data', None]}
            ]
        }
        if self.is_skip:
            api_items = list(api_definition.items())
            api_items.insert(1, ('skipIf', '${skip_test_in_production_env()}'))
            api_definition = dict(api_items)

        # 构建API引用用例
        api_cite = None
        if self.is_cite:
            api_cite = {
                'config': {
                    'name': f"{field_name} {operation_description}测试"
                },
                'teststeps': [
                    {
                        'name': f"执行 {field_name} {operation_description}",
                        'api': f'{output_dir}/{operation_type}_{field_name}_api.yml',
                        'extract': [{'data': 'content.data'}],
                        'validate': [
                            {'eq': ['status_code', 200]},
                            {'contains': ['headers.Content-Type', 'application/json']},
                            {'ne': ['content.data', None]}
                        ]
                    }
                ]
            }
            if self.is_skip:
                api_items = list(api_cite['teststeps'][0].items())
                api_items.insert(1, ('skipIf', '${skip_test_in_production_env()}'))
                api_cite['teststeps'][0] = dict(api_items)

        return api_definition, api_cite

    def _generate_variables_definition(self, variables: Dict[str, Any]) -> str:
        """生成变量定义字符串"""
        if not variables:
            return ""
            
        var_defs = []
        for var_name, var_value in variables.items():
            # 从参数信息中获取原始类型
            original_type = None
            for field_name, field_info in self.schema.root_fields.items():
                if field_info['args'] and var_name in field_info['args']:
                    original_type = field_info['args'][var_name]['type']
                    break
            
            # 始终使用原始类型定义，不要从值推断类型
            if original_type:
                var_defs.append(f"${var_name}: {original_type}")
            else:
                # 如果找不到原始类型，才使用推断类型
                var_type = self._get_graphql_type_for_value(var_value)
                var_defs.append(f"${var_name}: {var_type}")
            
        return f"({', '.join(var_defs)})"
    
    def _get_graphql_type_for_value(self, value: Any, original_type: str = None) -> str:
        """根据值获取GraphQL类型
        
        Args:
            value: 要判断类型的值
            original_type: 原始GraphQL类型定义（如果有）
        """
        # 如果提供了原始类型，优先使用原始类型
        if original_type:
            return original_type
            
        if isinstance(value, str):
            return "String"
        elif isinstance(value, int):
            return "Int"
        elif isinstance(value, float):
            return "Float"
        elif isinstance(value, bool):
            return "Boolean"
        elif value is None:
            return "String"
        elif isinstance(value, list):
            if value:
                item_type = self._get_graphql_type_for_value(value[0])
                return f"[{item_type}]"
            return "[String]"
        return "String"  # 默认为字符串类型
    
    def _generate_example_value(self, type_name: str) -> Any:
        """为给定类型生成示例值"""
        # 处理列表类型，提取内部类型
        orig_type_name = type_name
        is_list = False
        if type_name.startswith('[') and type_name.endswith(']'):
            is_list = True
            type_name = type_name[1:-1].rstrip('!')
        
        # 处理非空类型，移除'!'后缀
        is_non_null = False
        if type_name.endswith('!'):
            type_name = type_name[:-1]
            is_non_null = True
            
        # 为基本类型生成示例值
        if type_name == 'ID':
            # 使用统一的 ID 格式
            value = f"{random.randint(1, 1000)}"
        elif type_name == 'String':
            value = f"example-{random.randint(1, 1000)}"
        elif type_name == 'Int':
            value = random.randint(1, 100)
        elif type_name == 'Float':
            value = round(random.uniform(1.0, 100.0), 2)
        elif type_name == 'Boolean':
            value = random.choice([True, False])
        elif type_name == 'JSON':
            # 为JSON类型生成一个简单的JSON对象
            value = {"key": f"value-{random.randint(1, 100)}"}
        elif type_name == 'LocalDate':
            # 生成YYYY-MM-DD格式的日期
            today = datetime.date.today()
            delta = datetime.timedelta(days=random.randint(-30, 30))
            random_date = today + delta
            value = random_date.strftime("%Y-%m-%d")
        elif type_name == 'LocalDateTime':
            # 生成YYYY-MM-DDTHH:mm:ss格式的日期时间
            now = datetime.datetime.now()
            delta = datetime.timedelta(days=random.randint(-30, 30), 
                                    hours=random.randint(0, 23),
                                    minutes=random.randint(0, 59),
                                    seconds=random.randint(0, 59))
            random_datetime = now + delta
            value = random_datetime.strftime("%Y-%m-%dT%H:%M:%S")
        elif type_name == 'MixID':
            # 生成满足/^[\w-]*$/格式的ID
            chars = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-_"
            value = ''.join(random.choice(chars) for _ in range(random.randint(5, 10)))
        elif type_name == 'NumID' or type_name == 'NumStr':
            # 生成纯数字ID
            value = f"{random.randint(1, 1000)}"
        elif type_name == 'Void':
            # Void类型通常表示null
            value = None
        elif type_name in self.enum_types:
            # 为枚举类型生成一个合适的值
            # 如果有enum_values可用，从中随机选择一个
            if hasattr(self.schema, 'enum_values') and type_name in self.schema.enum_values:
                enum_values = list(self.schema.enum_values[type_name])
                if enum_values:
                    value = random.choice(enum_values)
                else:
                    # 如果没有定义enum值，生成一个占位符
                    value = f"{type_name.upper()}_VALUE_{random.randint(1, 10)}"
            else:
                # 如果没有定义enum值，生成一个占位符
                value = f"{type_name.upper()}_VALUE_{random.randint(1, 10)}"
        elif type_name.endswith('Type') and type_name not in self.schema.types:
            # 处理未在Schema中明确定义但以Type结尾的类型
            value = f"{type_name.lower()}-{random.randint(1, 100)}"
        else:
            # 如果是自定义类型但不在已知自定义标量列表中，使用默认ID格式
            if type_name in self.scalar_types:
                # 对于其他自定义标量类型，默认生成字符串
                value = f"custom-{type_name.lower()}-{random.randint(1, 1000)}"
            else:
                # 对于对象类型，返回纯数字格式的ID
                value = f"{random.randint(1, 1000)}"
        
        # 如果是列表类型，生成1-3个元素的列表
        if is_list:
            list_len = random.randint(1, 3)
            # 对于基本类型，生成多个不同的值
            if type_name in self.scalar_types or type_name in self.enum_types:
                return [self._generate_example_value(type_name) for _ in range(list_len)]
            # 对于对象类型，返回多个相同的值
            return [value] * list_len
                
        return value
    
    def _generate_fields_for_type(self, type_name: str, depth: int = 0, max_depth: int = None) -> str:
        """为类型生成查询字段"""
        # 使用实例的max_depth或传入的max_depth
        if max_depth is None:
            max_depth = self.max_depth
            
        # 防止无限递归
        if depth > max_depth:
            return "    id\n"
        
        # 如果是标量类型或枚举类型，直接返回空
        if type_name in self.scalar_types or type_name in self.schema.enum_types:
            return ""
        
        # 如果是列表类型，处理内部类型
        if type_name.startswith('[') and type_name.endswith(']'):
            inner_type = type_name[1:-1]
            return self._generate_fields_for_type(inner_type, depth, max_depth)
        
        # 检查类型是否存在
        if type_name not in self.schema.types:
            return "    id\n"
        
        # 防止重复生成相同类型
        if type_name in self.generated_types and depth > 0:
            return "    id\n"
        
        self.generated_types.add(type_name)
        
        # 获取类型对象
        type_obj = self.schema.types[type_name]
        
        # 构建字段字符串
        fields = []
        
        # 添加id字段（如果存在）
        if 'id' in type_obj.fields:
            fields.append("    id")
        
        # 添加其他标量字段和选择的复杂字段
        for field_name, field_info in type_obj.fields.items():
            if field_name == 'id':
                continue
                
            field_type = field_info['type']
            
            # 处理字段类型中的非空标记
            if field_type.endswith('!'):
                field_type = field_type[:-1]
                
            # 处理列表类型
            if field_type.startswith('[') and field_type.endswith(']'):
                field_type = field_type[1:-1]
                if field_type.endswith('!'):
                    field_type = field_type[:-1]
                
            # 判断是否应该包含该字段
            is_scalar = field_type in self.scalar_types
            is_enum = field_type in self.schema.enum_types
            is_type_suffix = field_type.endswith('Type')  # 处理以Type结尾的类型
            is_complex_type = field_type in self.schema.types
            has_no_args = not field_info['args']
            
            # 标量类型、枚举类型或以Type结尾的类型都直接添加
            if is_scalar or is_enum or (is_type_suffix and not is_complex_type):
                fields.append(f"    {field_name}")
            # 复杂类型需要递归处理
            elif has_no_args and is_complex_type and depth < max_depth:
                sub_fields = self._generate_fields_for_type(field_type, depth + 1, max_depth)
                if sub_fields:
                    fields.append(f"    {field_name} {{\n{sub_fields}    }}")
        
        # 确保至少有一个字段
        if not fields:
            if 'id' in type_obj.fields:
                fields.append("    id")
            else:
                # 如果连id字段都没有，选择第一个可用的标量字段
                for field_name, field_info in type_obj.fields.items():
                    field_type = field_info['type']
                    # 处理非空类型
                    if field_type.endswith('!'):
                        field_type = field_type[:-1]
                    if field_type in self.scalar_types or field_type in self.schema.enum_types:
                        fields.append(f"    {field_name}")
                        break
        
        # 移除类型，允许在其他分支中使用
        self.generated_types.remove(type_name)
        
        return "\n".join(fields) + "\n"


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
GraphQL 内省查询工具

该模块提供了通过内省查询(Introspection Query)获取GraphQL Schema的功能。
使用HTTP请求向GraphQL服务发送内省查询，获取完整的Schema信息，然后将其转换为
模型中的GraphQLSchema对象，供后续测试用例生成使用。

主要功能：
1. 向GraphQL服务发送内省查询请求
2. 解析内省查询结果
3. 将内省查询结果转换为GraphQLSchema对象
"""

import json
import requests
from typing import Dict, Any, Optional, List, Set

from .models import GraphQLType, GraphQLSchema


# GraphQL内省查询语句
INTROSPECTION_QUERY = """
query IntrospectionQuery {
  __schema {
    queryType {
      name
    }
    mutationType {
      name
    }
    subscriptionType {
      name
    }
    types {
      ...FullType
    }
    directives {
      name
      description
      locations
      args {
        ...InputValue
      }
    }
  }
}

fragment FullType on __Type {
  kind
  name
  description
  fields(includeDeprecated: true) {
    name
    description
    args {
      ...InputValue
    }
    type {
      ...TypeRef
    }
    isDeprecated
    deprecationReason
  }
  inputFields {
    ...InputValue
  }
  interfaces {
    ...TypeRef
  }
  enumValues(includeDeprecated: true) {
    name
    description
    isDeprecated
    deprecationReason
  }
  possibleTypes {
    ...TypeRef
  }
}

fragment InputValue on __InputValue {
  name
  description
  type {
    ...TypeRef
  }
  defaultValue
}

fragment TypeRef on __Type {
  kind
  name
  ofType {
    kind
    name
    ofType {
      kind
      name
      ofType {
        kind
        name
        ofType {
          kind
          name
          ofType {
            kind
            name
            ofType {
              kind
              name
              ofType {
                kind
                name
              }
            }
          }
        }
      }
    }
  }
}
"""


class IntrospectionQueryError(Exception):
    """内省查询错误"""
    pass


def fetch_schema_from_introspection(url: str) -> GraphQLSchema:
    """
    通过内省查询获取GraphQL Schema
    
    Args:
        url: GraphQL服务的URL地址
        
    Returns:
        GraphQLSchema: 解析后的Schema对象
        
    Raises:
        IntrospectionQueryError: 内省查询失败时抛出
    """
    print(f"正在从 {url} 进行内省查询获取GraphQL Schema...")
    
    # 发送内省查询请求
    try:
        response = requests.post(
            url,
            json={
                "query": INTROSPECTION_QUERY,
                "operationName": "IntrospectionQuery",
                "variables": {}
            },
            headers={"Content-Type": "application/json"}
        )
        response.raise_for_status()
    except requests.exceptions.RequestException as e:
        raise IntrospectionQueryError(f"内省查询请求失败: {str(e)}")
    
    # 解析响应内容
    try:
        result = response.json()
        # 检查是否有错误
        if "errors" in result:
            errors = result["errors"]
            error_msg = "; ".join([str(error.get("message", "未知错误")) for error in errors])
            raise IntrospectionQueryError(f"GraphQL服务返回错误: {error_msg}")
            
        # 确保结果中包含schema信息
        if "data" not in result or "__schema" not in result["data"]:
            raise IntrospectionQueryError("响应中未包含Schema信息")
            
        schema_data = result["data"]["__schema"]
        
    except (json.JSONDecodeError, KeyError) as e:
        raise IntrospectionQueryError(f"解析内省查询结果失败: {str(e)}")
    
    # 将内省查询结果转换为GraphQLSchema对象
    return _convert_introspection_to_schema(schema_data)


def _get_type_string(type_ref: Dict[str, Any]) -> str:
    """
    递归处理类型引用，生成与schema解析一致的类型字符串表示
    
    Args:
        type_ref: 类型引用数据
        
    Returns:
        str: 类型的字符串表示
    """
    kind = type_ref.get("kind")
    name = type_ref.get("name")
    
    # 基本命名类型(SCALAR, OBJECT, ENUM, etc.)
    if kind in ["SCALAR", "OBJECT", "INTERFACE", "UNION", "ENUM", "INPUT_OBJECT"] and name:
        return name
    # 列表类型
    elif kind == "LIST" and "ofType" in type_ref:
        inner_type = _get_type_string(type_ref["ofType"])
        return f"[{inner_type}]"
    # 非空类型
    elif kind == "NON_NULL" and "ofType" in type_ref:
        inner_type = _get_type_string(type_ref["ofType"])
        return f"{inner_type}!"
    # 未知类型
    return "Unknown"


def _parse_type_reference(type_ref: Dict[str, Any]) -> Dict[str, Any]:
    """
    解析类型引用，递归处理嵌套类型
    
    Args:
        type_ref: 类型引用数据
        
    Returns:
        Dict: 包含类型信息的字典
    """
    # 获取完整的类型字符串表示
    type_str = _get_type_string(type_ref)
    
    # 检查是否是列表类型
    is_list = False
    if type_str.startswith('[') and ']' in type_str:
        is_list = True
    
    # 检查是否是必需的(非空)
    required = False
    if type_str.endswith('!'):
        required = True
    
    # 提取基础类型名称
    # 对于列表类型[Type!]!，提取Type
    base_type = type_str
    if is_list:
        # 移除外层的[]
        base_type = type_str[1:type_str.rindex(']')]
        # 如果内部类型有非空标记，移除它
        if base_type.endswith('!'):
            base_type = base_type[:-1]
    # 如果类型有非空标记，移除它
    elif required:
        base_type = base_type[:-1]
    
    return {
        "type": base_type,        # 基础类型名称
        "type_str": type_str,     # 完整类型字符串表示
        "is_list": is_list,       # 是否是列表
        "required": required      # 是否是必需的
    }


def _update_arg_default_values(args: Dict[str, Dict[str, Any]], input_data: Dict[str, Any]) -> None:
    """
    从内省查询结果中提取参数的默认值
    
    Args:
        args: 参数字典，将被修改
        input_data: 包含defaultValue字段的内省数据
    """
    # 当defaultValue存在并且不为null时，记录默认值信息
    if input_data.get("defaultValue") is not None:
        arg_name = input_data.get("name")
        if arg_name in args:
            # 处理defaultValue，可能需要解析JSON字符串
            default_value = input_data.get("defaultValue")
            try:
                # 尝试解析JSON字符串
                if default_value and (default_value.startswith('{') or default_value.startswith('[')):
                    import json
                    default_value = json.loads(default_value)
            except:
                # 解析失败则保留原始字符串
                pass
                
            args[arg_name]["default_value"] = default_value


def _convert_introspection_to_schema(introspection_data: Dict[str, Any]) -> GraphQLSchema:
    """
    将内省查询结果转换为GraphQLSchema对象
    
    Args:
        introspection_data: 内省查询返回的Schema数据
        
    Returns:
        GraphQLSchema: 转换后的Schema对象
    """
    print("正在解析内省查询结果...")
    
    schema = GraphQLSchema()
    
    # 设置根类型
    if introspection_data.get("queryType") and introspection_data["queryType"].get("name"):
        schema.set_query_type(introspection_data["queryType"]["name"])
        
    if introspection_data.get("mutationType") and introspection_data["mutationType"].get("name"):
        schema.set_mutation_type(introspection_data["mutationType"]["name"])
        
    if introspection_data.get("subscriptionType") and introspection_data["subscriptionType"].get("name"):
        schema.set_subscription_type(introspection_data["subscriptionType"]["name"])
    
    # 处理所有类型
    type_map = {}  # 用于临时存储类型信息
    
    for type_data in introspection_data.get("types", []):
        kind = type_data.get("kind")
        name = type_data.get("name")
        
        if not name:  # 跳过没有名称的类型
            continue
            
        # 跳过内置类型
        if name.startswith("__"):
            continue
            
        # 处理SCALAR类型
        if kind == "SCALAR" and name not in ["String", "Int", "Float", "Boolean", "ID"]:
            schema.add_scalar_type(name)
            
        # 处理ENUM类型
        elif kind == "ENUM":
            schema.add_enum_type(name)
            enum_values = set()
            for enum_value in type_data.get("enumValues", []):
                enum_values.add(enum_value["name"])
            schema.add_enum_values(name, enum_values)
            
        # 处理OBJECT类型
        elif kind == "OBJECT":
            # 创建GraphQLType对象
            type_obj = GraphQLType(
                name=name,
                description=type_data.get("description")
            )
            
            # 解析实现的接口
            interfaces = []
            for interface in type_data.get("interfaces", []):
                if interface.get("name"):
                    interfaces.append(interface["name"])
            type_obj.implements = interfaces
            
            # 解析字段
            for field in type_data.get("fields", []):
                field_name = field["name"]
                type_info = _parse_type_reference(field["type"])
                
                # 解析参数
                args = {}
                for arg in field.get("args", []):
                    arg_name = arg["name"]
                    arg_type_info = _parse_type_reference(arg["type"])
                    
                    # 使用完整的类型字符串作为类型
                    args[arg_name] = {
                        "type": arg_type_info["type_str"],
                        "required": arg_type_info["required"],
                        "is_list": arg_type_info["is_list"]
                    }
                    
                    # 提取默认值
                    _update_arg_default_values(args, arg)
                
                # 添加字段信息
                field_info = {
                    "type": type_info["type_str"],  # 使用完整类型字符串
                    "is_list": type_info["is_list"],
                    "required": type_info["required"],
                    "args": args
                }
                type_obj.add_field(field_name, field_info)
            
            # 添加到Schema
            schema.add_type(type_obj)
            
            # 记录类型信息，用于后续处理根字段
            type_map[name] = type_obj
    
    # 添加根字段
    query_type_name = schema.query_type
    if query_type_name and query_type_name in type_map:
        for field_name, field_info in type_map[query_type_name].fields.items():
            schema.add_root_field(field_name, field_info)
    
    mutation_type_name = schema.mutation_type
    if mutation_type_name and mutation_type_name in type_map:
        for field_name, field_info in type_map[mutation_type_name].fields.items():
            schema.add_root_field(field_name, field_info)
    
    subscription_type_name = schema.subscription_type
    if subscription_type_name and subscription_type_name in type_map:
        for field_name, field_info in type_map[subscription_type_name].fields.items():
            schema.add_root_field(field_name, field_info)
    
    print(f"Schema解析完成，包含 {len(schema.types)} 个类型和 {len(schema.root_fields)} 个根字段")
    
    return schema 


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
GraphQL Schema to HttpRunner 工具类模块

此模块提供工具函数，主要包括：
1. 查询语句文件备份功能
2. 查询语句文件差异对比与报告生成功能
3. 测试用例自动更新功能

这些功能主要用于在生成新的GraphQL查询语句时，保留旧版本并生成详细的差异比较报告，
便于用户了解API变更情况，实现API变更的可视化跟踪，并自动更新维护测试用例。
"""

import os
import shutil
import datetime
import yaml
import glob
import csv

from .report_generator import generate_markdown_diff_report, generate_html_diff_report


def backup_queries_file(file_path):
    """
    备份查询语句文件
    
    将现有的查询语句文件备份为带时间戳的文件，并删除原文件
    
    Args:
        file_path (str): 需要备份的文件路径
        
    Returns:
        str or None: 备份文件的路径，失败时返回None
    """
    if not os.path.exists(file_path):
        return None
    
    # 分离文件名和扩展名
    file_name, file_ext = os.path.splitext(file_path)
    backup_path = f"{file_name}.bak.{datetime.datetime.now().strftime('%Y%m%d%H%M%S')}{file_ext}"
    
    try:
        shutil.copy2(file_path, backup_path)
        print(f"备份旧的查询语句文件: {file_path} -> {backup_path}")
        os.remove(file_path)
        print(f"删除旧的查询语句文件: {file_path}")
        return backup_path
    except Exception as e:
        print(f"备份旧查询语句文件时出错: {e}")
        return None


def compare_query_files(old_file, new_file, report_format=None):
    """
    比较新旧查询语句文件的差异并保存差异结果
    
    分析新旧查询语句文件的差异，根据指定格式生成差异报告，
    包括项目级变更、查询级变更以及详细的文本差异
    
    Args:
        old_file (str): 旧查询语句文件路径
        new_file (str): 新查询语句文件路径
        report_format (str): 报告格式，可选值：'markdown'、'html'、'both'，默认为None，即不生成差异报告
        
    Returns:
        dict or True or False: 存在差异，返回差异字典信息；无差异时，返回True；出现错误返回False
    """
    if not os.path.exists(old_file) or not os.path.exists(new_file):
        print(f"无法进行文件对比，旧文件或新文件不存在")
        return False
    
    # 读取旧文件内容
    try:
        with open(old_file, 'r', encoding='utf-8') as f:
            old_content = yaml.safe_load(f)
    except Exception as e:
        print(f"读取旧文件内容时出错: {e}")
        return False
    
    # 读取新文件内容
    try:
        with open(new_file, 'r', encoding='utf-8') as f:
            new_content = yaml.safe_load(f)
    except Exception as e:
        print(f"读取新文件内容时出错: {e}")
        return False
    
    # 分析差异
    diff_result = {
        "added_projects": [],
        "removed_projects": [],
        "modified_projects": {},
        "added_queries": {},
        "removed_queries": {},
        "modified_queries": {},
    }
    
    # 检查项目级差异
    old_projects = set(old_content.keys())
    new_projects = set(new_content.keys())
    
    diff_result["added_projects"] = list(new_projects - old_projects)
    diff_result["removed_projects"] = list(old_projects - new_projects)
    
    # 对于共同存在的项目，检查查询语句差异
    common_projects = old_projects.intersection(new_projects)
    for project in common_projects:
        old_queries = set(old_content[project].keys())
        new_queries = set(new_content[project].keys())
        
        # 检查是否有查询语句差异
        added_queries = new_queries - old_queries
        if added_queries:
            diff_result["added_queries"][project] = list(added_queries)
        
        removed_queries = old_queries - new_queries
        if removed_queries:
            diff_result["removed_queries"][project] = list(removed_queries)
        
        # 检查共同查询语句的内容差异
        common_queries = old_queries.intersection(new_queries)
        modified_queries = []
        
        for query in common_queries:
            if old_content[project][query] != new_content[project][query]:
                modified_queries.append(query)
        
        if modified_queries:
            diff_result["modified_queries"][project] = modified_queries
        
        # 如果项目有任何变化，记录到modified_projects
        if (project in diff_result["added_queries"] or 
            project in diff_result["removed_queries"] or 
            project in diff_result["modified_queries"]):
            diff_result["modified_projects"][project] = {
                "added_queries": diff_result["added_queries"].get(project, []),
                "removed_queries": diff_result["removed_queries"].get(project, []),
                "modified_queries": diff_result["modified_queries"].get(project, [])
            }
    
    # 判断是否存在差异
    has_difference = (diff_result["added_projects"] or 
                     diff_result["removed_projects"] or 
                     diff_result["modified_projects"])
    
    if has_difference:
        # 生成时间戳
        timestamp = datetime.datetime.now().strftime('%Y%m%d%H%M%S')
        
        # 根据指定格式生成差异报告
        if report_format in ['markdown', 'both']:
            # 生成并保存Markdown格式差异结果报告
            md_diff_file = f"{new_file}.{timestamp}.diff.md"
            try:
                md_content = generate_markdown_diff_report(diff_result, old_content, new_content, old_file, new_file)
                with open(md_diff_file, 'w', encoding='utf-8') as f:
                    f.write('\n'.join(md_content))
                print(f"生成查询语句差异报告(Markdown格式): {md_diff_file}")
            except Exception as e:
                print(f"保存查询语句差异报告(Markdown格式)时出错: {e}")
                return False
        
        if report_format in ['html', 'both']:
            # 生成并保存HTML格式差异结果
            html_diff_file = f"{new_file}.{timestamp}.diff.html"
            try:
                html_content = generate_html_diff_report(diff_result, old_content, new_content, old_file, new_file)
                with open(html_diff_file, 'w', encoding='utf-8') as f:
                    f.write(html_content)
                print(f"生成查询语句差异报告(HTML格式): {html_diff_file}")
            except Exception as e:
                print(f"保存查询语句差异报告(HTML格式)时出错: {e}")
                return False
        
        # 输出汇总信息
        print(f"\n新旧查询语句文件差异比较结果:")
        if diff_result["added_projects"]:
            print(f"- 新增项目: {len(diff_result['added_projects'])}")
        if diff_result["removed_projects"]:
            print(f"- 移除项目: {len(diff_result['removed_projects'])}")
        
        modified_count = len(diff_result["modified_projects"])
        if modified_count:
            print(f"- 修改项目: {modified_count}")
            
            added_queries_count = sum(len(queries) for queries in diff_result["added_queries"].values())
            if added_queries_count:
                print(f"  - 新增查询: {added_queries_count}")
            
            removed_queries_count = sum(len(queries) for queries in diff_result["removed_queries"].values())
            if removed_queries_count:
                print(f"  - 移除查询: {removed_queries_count}")
            
            modified_queries_count = sum(len(queries) for queries in diff_result["modified_queries"].values())
            if modified_queries_count:
                print(f"  - 修改查询: {modified_queries_count}")
        
        print("详细差异报告已生成成功！")
        # 返回差异结果
        return diff_result
    else:
        print(f"\n查询语句文件比较结果: 没有发现差异！")
        return True


def find_testcase_files(output_dir, operation_type, query_name, is_api=False):
    """
    查找与指定查询对应的测试用例文件
    
    Args:
        output_dir (str): 输出目录路径
        operation_type (str): 操作类型（query或mutation）
        query_name (str): 查询名称
        is_api (bool): 是否查找API层测试用例
        
    Returns:
        list: 匹配的测试用例文件路径列表
    """
    file_suffix = "_api.yml" if is_api else "_test.yml"
    pattern = os.path.join(output_dir, "**", f"{operation_type}_{query_name}{file_suffix}")
    return glob.glob(pattern, recursive=True)


def delete_testcase_files(output_dir, operation_type, query_name, is_api=False):
    """
    删除与指定查询对应的测试用例文件
    
    Args:
        output_dir (str): 输出目录路径
        operation_type (str): 操作类型（query或mutation）
        query_name (str): 查询名称
        is_api (bool): 是否删除API层测试用例
        
    Returns:
        int: 删除的文件数量
    """
    files = find_testcase_files(output_dir, operation_type, query_name, is_api)
    deleted_count = 0
    
    for file_path in files:
        try:
            os.remove(file_path)
            print(f"删除测试用例文件: {file_path}")
            deleted_count += 1
        except Exception as e:
            print(f"删除测试用例文件失败: {file_path}, 错误: {e}")
    
    return deleted_count


def update_testcase_query(file_path, new_query):
    """
    更新测试用例文件中的查询语句
    
    Args:
        file_path (str): 测试用例文件路径
        new_query (str): 新的查询语句
        
    Returns:
        bool: 更新是否成功
    """
    try:
        # 读取测试用例文件
        with open(file_path, 'r', encoding='utf-8') as f:
            testcase = yaml.safe_load(f)
        
        # 更新查询语句
        if 'teststeps' in testcase and isinstance(testcase['teststeps'], list) and len(testcase['teststeps']) > 0:
            # 用例层测试用例
            if 'request' in testcase['teststeps'][0] and 'json' in testcase['teststeps'][0]['request']:
                json_data = testcase['teststeps'][0]['request']['json']
                if 'query' in json_data:
                    # 转义GraphQL查询中的$符号，避免与HttpRunner变量语法冲突
                    json_data['query'] = new_query.replace("$", "$$")
        elif 'request' in testcase and 'json' in testcase['request']:
            # API层测试用例
            json_data = testcase['request']['json']
            if 'query' in json_data:
                # 转义GraphQL查询中的$符号，避免与HttpRunner变量语法冲突
                json_data['query'] = new_query.replace("$", "$$")
        
        # 保存更新后的测试用例
        with open(file_path, 'w', encoding='utf-8') as f:
            yaml.dump(testcase, f, default_flow_style=False, sort_keys=False, allow_unicode=True, width=9999)
        
        print(f"更新测试用例文件查询语句: {file_path}")
        return True
    except Exception as e:
        print(f"更新测试用例文件查询语句失败: {file_path}, 错误: {e}")
        return False


def determine_operation_type(query):
    """
    根据查询语句确定操作类型
    
    Args:
        query (str): 查询语句
        
    Returns:
        str: 操作类型（"query"或"mutation"）
    """
    # 检查查询语句是否以mutation开头
    if query.strip().startswith("mutation"):
        return "mutation"
    return "query"


def auto_update_testcases(diff_result, config_file, queries_file):
    """
    根据差异报告自动更新测试用例
    
    Args:
        diff_result (dict): 差异比较结果
        config_file (str): 配置文件路径
        queries_file (str): 查询语句文件路径
        
    Returns:
        dict: 更新统计结果
    """
    if not diff_result:
        print("没有变更需要处理")
        return {}
    
    # 读取配置文件
    projects_config = {}
    try:
        with open(config_file, "r", encoding="utf-8") as f:
            reader = csv.DictReader(f)
            for row in reader:
                projects_config[row["project_name"]] = row
    except Exception as e:
        print(f"读取配置文件时出错: {e}")
        return {}
    
    # 读取查询语句文件
    try:
        with open(queries_file, 'r', encoding='utf-8') as f:
            queries_content = yaml.safe_load(f)
    except Exception as e:
        print(f"读取查询语句文件时出错: {e}")
        return {}
    
    # 更新统计
    stats = {
        "new_projects": 0,
        "removed_projects": 0,
        "added_queries": 0,
        "removed_queries": 0,
        "modified_queries": 0
    }
    
    print(f"\n{'='*80}")
    print("开始自动更新测试用例:")
    
    # 处理新增项目
    for project in diff_result.get("added_projects", []):
        if project in projects_config:
            config = projects_config[project]
            output_dir = config["output"]
            introspection_url = config["introspection_url"]
            base_url = config["base_url"]
            max_depth = int(config.get("max_depth", 2))
            required = config.get("is_required", "false").lower() == "true"
            is_api = config.get("is_api", "false").lower() == "true"
            is_cite = config.get("is_cite", "false").lower() == "true"
            is_skip = config.get("is_skip", "false").lower() == "true"
            
            print(f"\n处理新增项目: {project}")
            
            # 如果输出目录不存在，则创建
            if not os.path.exists(output_dir):
                os.makedirs(output_dir, exist_ok=True)
            
            # 导入process_single_project函数
            from .main import process_single_project
            
            # 生成新项目的测试用例
            result = process_single_project(
                introspection_url=introspection_url,
                output=output_dir,
                base_url=base_url,
                max_depth=max_depth,
                is_api=is_api,
                is_cite=is_cite,
                is_skip=is_skip,
                required=required,
                is_testcases=True,
                project_name=project
            )
            
            if result:
                stats["new_projects"] += 1
                print(f"新增项目 {project} 的测试用例生成成功")
            else:
                print(f"新增项目 {project} 的测试用例生成失败")
        else:
            print(f"新增项目 {project} 在配置文件中未找到，跳过处理")
    
    # 处理移除项目
    for project in diff_result.get("removed_projects", []):
        if project in projects_config:
            config = projects_config[project]
            output_dir = config["output"]
            is_api = config.get("is_api", "false").lower() == "true"
            is_cite = config.get("is_cite", "false").lower() == "true"
            
            if is_api and is_cite:
                # 如果是API，并且有引用层，则需要删除引用层测试用例文件夹
                cite_output_dir = output_dir.replace("api", "testcases", 1) if 'api' in output_dir else output_dir

            if os.path.exists(output_dir):
                try:
                    print(f"\n处理移除项目: {project}")
                    
                    # 删除项目目录
                    shutil.rmtree(output_dir)
                    stats["removed_projects"] += 1
                    print(f"移除项目 {project} 的测试用例目录: {output_dir}")

                    # 删除引用层测试用例目录
                    if cite_output_dir and os.path.exists(cite_output_dir):
                        # 如果引用层测试用例目录与项目目录相同，则前面已删除了
                        if cite_output_dir != output_dir:
                            shutil.rmtree(cite_output_dir)
                        print(f"移除项目 {project} 的引用层测试用例目录: {cite_output_dir}")
                    else:
                        print(f"注意：移除项目 {project} 的引用层测试用例目录不存在，跳过处理")
                except Exception as e:
                    print(f"移除项目 {project} 的测试用例目录失败: {e}")
            else:
                print(f"注意：移除项目 {project} 的测试用例目录不存在，跳过处理")
        else:
            print(f"注意：移除项目 {project} 在配置文件中未找到，跳过处理")
    
    # 处理修改项目
    for project, changes in diff_result.get("modified_projects", {}).items():
        if project in projects_config:
            config = projects_config[project]
            output_dir = config["output"]
            is_api = config.get("is_api", "false").lower() == "true"
            is_cite = config.get("is_cite", "false").lower() == "true"
            
            print(f"\n处理修改项目: {project}")
            
            # 处理移除的查询
            for query in changes.get("removed_queries", []):
                # 确定操作类型
                if project in queries_content and query in queries_content[project]:
                    print(f"ERROR: 查询 {query} 在新的查询语句中存在，但在差异报告中被标记为移除！")
                    continue
                
                # 由于旧查询已不存在，无法确定操作类型，尝试删除query和mutation两种类型的文件
                deleted_count = delete_testcase_files(output_dir, "query", query, is_api)
                deleted_count += delete_testcase_files(output_dir, "mutation", query, is_api)
                if is_api and is_cite:
                    cite_output_dir = output_dir.replace("api", "testcases", 1) if 'api' in output_dir else output_dir
                    cite_deleted_count = delete_testcase_files(cite_output_dir, "query", query, False)
                    cite_deleted_count += delete_testcase_files(cite_output_dir, "mutation", query, False)
                
                if deleted_count > 0:
                    stats["removed_queries"] += 1
                    print(f"移除查询 {query} 的测试用例文件，共 {deleted_count} 个")
                    if cite_deleted_count > 0:
                        print(f"移除查询 {query} 的引用层测试用例文件，共 {cite_deleted_count} 个")
            
            # 处理新增的查询
            added_queries = changes.get("added_queries", [])
            if added_queries:
                # 导入需要的模块
                from .introspection import fetch_schema_from_introspection
                from .generator import HttpRunnerTestCaseGenerator
                try:
                    introspection_url = config["introspection_url"]
                    base_url = config["base_url"]
                    max_depth = int(config.get("max_depth", 2))
                    required = config.get("is_required", "false").lower() == "true"
                    is_skip = config.get("is_skip", "false").lower() == "true"
                    is_cite = config.get("is_cite", "false").lower() == "true"
                    # 获取GraphQL Schema
                    schema = fetch_schema_from_introspection(introspection_url)
                    # 创建生成器
                    generator = HttpRunnerTestCaseGenerator(schema, base_url, max_depth, required, is_skip, is_cite)
                    # 生成新增查询的测试用例
                    for query in added_queries:
                        if is_api:
                            result = generator.generate_single_api_test_case(output_dir, query)
                        else:
                            result = generator.generate_single_test_case(output_dir, query)

                        if result == 1:
                            stats["added_queries"] += 1
                            print(f"新增查询 {query} 的测试用例文件生成成功")
                        else:
                            print(f"新增查询 {query} 的测试用例文件生成失败")
                except Exception as e:
                    print(f"获取{project}项目的GraphQL Schema失败: {e}，无法生成新增查询的测试用例：{added_queries}")
                    continue

            # 处理修改的查询
            for query in changes.get("modified_queries", []):
                if project in queries_content and query in queries_content[project]:
                    query_text = queries_content[project][query]
                    operation_type = determine_operation_type(query_text)
                    
                    print(f"处理修改查询: {query}")
                    
                    # 查找对应的测试用例文件
                    files = find_testcase_files(output_dir, operation_type, query, is_api)
                    
                    if files:
                        for file_path in files:
                            # 更新测试用例中的查询语句
                            if update_testcase_query(file_path, query_text):
                                stats["modified_queries"] += 1
                    else:
                        print(f"注意：未找到修改查询 {query} 对应的测试用例文件，跳过")
                        # # 未找到修改查询，作为新增查询处理
                        # from .introspection import fetch_schema_from_introspection
                        # from .generator import HttpRunnerTestCaseGenerator
                        # try:
                        #     introspection_url = config["introspection_url"]
                        #     base_url = config["base_url"]
                        #     max_depth = int(config.get("max_depth", 2))
                        #     required = config.get("is_required", "false").lower() == "true"
                        #     is_skip = config.get("is_skip", "false").lower() == "true"
                        #     is_cite = config.get("is_cite", "false").lower() == "true"
                        #     # 获取GraphQL Schema
                        #     schema = fetch_schema_from_introspection(introspection_url)
                        #     # 创建生成器
                        #     generator = HttpRunnerTestCaseGenerator(schema, base_url, max_depth, required, is_skip, is_cite)
                        #     result = generator.generate_single_test_case(output_dir, query)
                        #     if result == 1:
                        #         stats["added_queries"] += 1
                        #         print(f"修改查询 {query} 未找到原测试用例，直接创建新用例成功")
                        #     else:
                        #         print(f"修改查询 {query} 未找到原测试用例，直接创建新用例失败")
                        # except Exception as e:
                        #     print(f"生成修改查询 {query} 的测试用例失败: {e}")
        else:
            print(f"修改项目 {project} 在配置文件中未找到，跳过处理")
    
    # 输出更新统计
    print(f"\n{'='*80}")
    print("自动更新测试用例完成:")
    print(f"- 新增项目: {stats['new_projects']}")
    print(f"- 移除项目: {stats['removed_projects']}")
    print(f"- 新增查询: {stats['added_queries']}")
    print(f"- 移除查询: {stats['removed_queries']}")
    print(f"- 修改查询: {stats['modified_queries']}")
    
    return stats


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
GraphQL Schema 解析器

该模块提供了GraphQL Schema文件的解析功能，将schema.graphql文件内容解析为
GraphQLSchema对象。支持解析类型定义、接口、枚举、标量等GraphQL Schema元素。

主要功能：
1. 解析GraphQL类型定义
2. 解析GraphQL接口定义
3. 解析GraphQL枚举和标量类型
4. 提取类型和字段的描述信息
5. 识别根查询、变更和订阅类型
"""

import re
from typing import Dict, Any, Optional, Set

from .models import GraphQLType, GraphQLSchema


class GraphQLSchemaParser:
    """GraphQL Schema解析器"""
    
    def __init__(self, schema_content: str):
        self.schema_content = schema_content
        self.schema = GraphQLSchema()
        
    def parse(self) -> GraphQLSchema:
        """解析GraphQL Schema"""
        # 解析自定义标量类型
        self._parse_scalar_types()
        
        # 解析枚举类型
        self._parse_enum_types()
        
        # 解析schema定义
        schema_match = re.search(r'schema\s*{([^}]*)}', self.schema_content)
        if schema_match:
            schema_def = schema_match.group(1)
            query_match = re.search(r'query:\s*(\w+)', schema_def)
            if query_match:
                self.schema.set_query_type(query_match.group(1))
            
            mutation_match = re.search(r'mutation:\s*(\w+)', schema_def)
            if mutation_match:
                self.schema.set_mutation_type(mutation_match.group(1))
                
            subscription_match = re.search(r'subscription:\s*(\w+)', schema_def)
            if subscription_match:
                self.schema.set_subscription_type(subscription_match.group(1))
        else:
            # Apollo GraphQL可能没有显式的schema定义，尝试找到Query类型
            print("未找到schema定义，尝试寻找Query类型...")
            query_type_match = re.search(r'type\s+Query\s*{', self.schema_content)
            if query_type_match:
                print("找到Query类型定义，设置为根查询类型")
                self.schema.set_query_type('Query')
            else:
                print("警告：未找到Query类型定义")

            mutation_type_match = re.search(r'type\s+Mutation\s*{', self.schema_content)
            if mutation_type_match:
                print("找到Mutation类型定义，设置为根变更类型")
                self.schema.set_mutation_type('Mutation')
                
            subscription_type_match = re.search(r'type\s+Subscription\s*{', self.schema_content)
            if subscription_type_match:
                print("找到Subscription类型定义，设置为根订阅类型")
                self.schema.set_subscription_type('Subscription')
        
        # 解析类型定义
        type_pattern = r'type\s+(\w+)(?:\s+implements\s+([^{]+))?\s*{([^}]*)}'
        for match in re.finditer(type_pattern, self.schema_content):
            type_name = match.group(1)
            implements_str = match.group(2) or ""
            fields_str = match.group(3)
            
            # 获取类型描述
            description = self._get_description_before(match.start())
            
            # 解析implements
            implements = [i.strip() for i in implements_str.split('&') if i.strip()]
            
            # 创建类型对象
            type_obj = GraphQLType(type_name, implements=implements, description=description)
            
            # 解析字段
            self._parse_fields(fields_str, type_obj)
            
            # 添加到schema
            self.schema.add_type(type_obj)
            
            # 如果是Root类型，解析根字段
            if type_name == self.schema.query_type:
                # 提取Root类型中直接定义的字段（不包括参数）
                root_fields = self._extract_top_level_fields(fields_str)
                
                for field_name, field_info in type_obj.fields.items():
                    # 只有在顶级字段列表中的字段才被添加为root字段
                    if field_name in root_fields:
                        self.schema.add_root_field(field_name, field_info)
                    else:
                        print(f"字段 {field_name} 不在顶级字段列表中，被忽略")
                
                # 打印最终的root_fields
                print(f"查询类型字段: {list(self.schema.root_fields.keys())}")
            
            # 如果是Mutation类型，也解析其字段
            elif type_name == self.schema.mutation_type:
                # 提取Mutation类型中直接定义的字段
                mutation_fields = self._extract_top_level_fields(fields_str)
                
                for field_name, field_info in type_obj.fields.items():
                    if field_name in mutation_fields:
                        # 将Mutation字段也添加到root_fields中
                        self.schema.add_root_field(field_name, field_info)
                    else:
                        print(f"字段 {field_name} 不在Mutation字段列表中，被忽略")
                
                print(f"添加变更类型字段: {list(self.schema.root_fields.keys())}")
            
            # 同样处理Subscription类型
            elif type_name == self.schema.subscription_type:
                # 提取Subscription类型中直接定义的字段
                subscription_fields = self._extract_top_level_fields(fields_str)
                
                for field_name, field_info in type_obj.fields.items():
                    if field_name in subscription_fields:
                        # 将Subscription字段也添加到root_fields中
                        self.schema.add_root_field(field_name, field_info)
                    else:
                        print(f"字段 {field_name} 不在Subscription字段列表中，被忽略")
                
                print(f"添加订阅类型字段: {list(self.schema.root_fields.keys())}")
        
        # 解析接口定义
        interface_pattern = r'interface\s+(\w+)\s*{([^}]*)}'
        for match in re.finditer(interface_pattern, self.schema_content):
            interface_name = match.group(1)
            fields_str = match.group(2)
            
            # 获取接口描述
            description = self._get_description_before(match.start())
            
            # 创建接口对象
            interface_obj = GraphQLType(interface_name, description=description)
            
            # 解析字段
            self._parse_fields(fields_str, interface_obj)
            
            # 添加到schema
            self.schema.add_interface(interface_obj)
        
        # 打印识别到的自定义标量类型和枚举类型
        print(f"识别到的标量类型: {self.schema.scalar_types}")
        print(f"识别到的枚举类型: {self.schema.enum_types}")
        
        return self.schema
        
    def _parse_scalar_types(self):
        """解析自定义标量类型定义"""
        scalar_pattern = r'scalar\s+(\w+)'
        for match in re.finditer(scalar_pattern, self.schema_content):
            scalar_name = match.group(1)
            self.schema.add_scalar_type(scalar_name)
            
    def _parse_enum_types(self):
        """解析枚举类型定义"""
        enum_pattern = r'enum\s+(\w+)\s*{([^}]*)}'
        for match in re.finditer(enum_pattern, self.schema_content):
            enum_name = match.group(1)
            self.schema.add_enum_type(enum_name)
            
            # 解析枚举值
            enum_values = [value.strip() for value in match.group(2).split('\n') if value.strip()]
            self.schema.add_enum_values(enum_name, enum_values)
    
    def _get_description_before(self, pos: int) -> Optional[str]:
        """获取位置前的描述注释"""
        # 向前查找最近的三引号注释
        content_before = self.schema_content[:pos].strip()
        desc_match = re.search(r'"""([^"]*)"""s*$', content_before, re.DOTALL)
        if desc_match:
            return desc_match.group(1).strip()
        return None
    
    def _parse_fields(self, fields_str: str, type_obj: GraphQLType):
        """解析字段定义"""
        # 预处理字段字符串，移除注释干扰
        cleaned_fields = []
        in_comment = False
        for line in fields_str.split('\n'):
            line = line.strip()
            if not line:
                continue
                
            # 处理注释行
            if '"""' in line:
                # 检查是否是注释的开始或结束
                if line.count('"""') % 2 == 1:  # 奇数个引号，表示开始或结束注释
                    in_comment = not in_comment
                
                # 提取行中的非注释部分
                non_comment_parts = line.split('"""')
                if len(non_comment_parts) > 2:  # 有开始和结束引号
                    clean_line = non_comment_parts[0] + non_comment_parts[-1]
                    if clean_line.strip():
                        cleaned_fields.append(clean_line.strip())
                continue
                
            # 跳过完整的注释行
            if in_comment:
                continue
                
            cleaned_fields.append(line)
        
        # 合并处理后的字段字符串，便于多行字段的解析
        processed_fields_str = '\n'.join(cleaned_fields)
        
        # 使用正则表达式匹配所有字段定义
        # 字段模式：字段名(参数): 类型
        field_pattern = r'(\w+)(?:\s*\(([^)]*)\))?\s*:\s*([^\s,{]+)(!?)'
        
        # 查找所有匹配的字段
        for match in re.finditer(field_pattern, processed_fields_str):
            field_name = match.group(1)
            args_str = match.group(2) or ""
            field_type = match.group(3)
            required = bool(match.group(4))
            
            # 解析参数
            args = {}
            if args_str:
                args_pattern = r'(\w+)\s*:\s*([^\s,]+)(!?)'
                for arg_match in re.finditer(args_pattern, args_str):
                    arg_name = arg_match.group(1)
                    arg_type = arg_match.group(2)
                    arg_required = bool(arg_match.group(3))
                    args[arg_name] = {
                        'type': arg_type,
                        'required': arg_required
                    }
            
            # 检查是否是列表类型
            is_list = False
            if field_type.startswith('[') and field_type.endswith(']'):
                is_list = True
                field_type = field_type[1:-1]
            
            # 添加字段
            type_obj.add_field(field_name, {
                'type': field_type,
                'required': required,
                'is_list': is_list,
                'args': args,
                'description': None  # 简化处理，不保留描述
            })

    def _extract_top_level_fields(self, fields_str: str) -> Set[str]:
        """
        提取字段定义字符串中的顶级字段名称
        
        Args:
            fields_str: 字段定义字符串
            
        Returns:
            顶级字段名称集合
        """
        # 将字段字符串按行分割
        lines = [line.strip() for line in fields_str.split('\n')]
        
        # 存储顶级字段名称
        top_level_fields = set()
        
        # 当前正在处理的字段
        current_field = None
        in_field_def = False
        in_comment = False
        open_parens = 0
        
        # 逐行处理，识别顶级字段定义
        for line in lines:
            # 跳过空行
            if not line:
                continue
                
            # 处理注释
            if line.startswith('"""'):
                # 如果是注释的开始或结束
                if line.count('"""') % 2 == 1:  # 奇数个引号，表示开始或结束注释
                    in_comment = not in_comment
                continue
                
            # 在注释中，跳过处理
            if in_comment:
                continue
                
            # 检查是否是字段定义的开始（在行首，不缩进）
            if not in_field_def:
                # 字段定义模式: 字段名(可能有参数): 类型
                field_match = re.match(r'^(\w+)', line)
                if field_match and not line.startswith(' '):
                    current_field = field_match.group(1)
                    in_field_def = True
                    
                    # 计算括号数量 - 帮助识别多行参数定义
                    open_parens += line.count('(') - line.count(')')
                    
                    # 如果这行包含冒号且括号已关闭，表示字段定义完成
                    if ':' in line and open_parens == 0:
                        top_level_fields.add(current_field)
                        in_field_def = False
                        current_field = None
            
            # 如果在字段定义过程中，继续处理直到定义结束
            elif in_field_def:
                # 更新括号数量
                open_parens += line.count('(') - line.count(')')
                
                # 如果找到冒号且括号已平衡，那么字段定义结束
                if ':' in line and open_parens <= 0:
                    top_level_fields.add(current_field)
                    in_field_def = False
                    current_field = None
        
        # 创建一个全局字段模式匹配，用于捕获可能被上面方法遗漏的字段
        # 特别是对于有多行复杂定义和注释的字段
        field_pattern = r'^\s*(\w+)\s*\('
        
        # 使用正则表达式查找所有可能的字段定义
        raw_content = '\n'.join(line for line in fields_str.split('\n') if line.strip())
        for match in re.finditer(field_pattern, raw_content, re.MULTILINE):
            top_level_fields.add(match.group(1))
        
        return top_level_fields 


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
GraphQL Schema 数据模型

该模块定义了GraphQL Schema的核心数据模型，包括GraphQLType和GraphQLSchema两个类，
用于表示GraphQL Schema的基本结构和类型信息。这些模型被GraphQLSchemaParser使用，
用于解析GraphQL Schema文件，并为测试用例生成提供基础数据结构。

主要包括：
1. GraphQLType - 表示GraphQL中的类型，包含类型名称、字段、实现接口等信息
2. GraphQLSchema - 表示完整的GraphQL Schema，包含所有类型、接口、查询、变更等信息
"""

from typing import Dict, List, Any, Optional, Set


class GraphQLType:
    """表示GraphQL类型的类"""
    
    def __init__(self, name: str, fields: Dict[str, Dict[str, Any]] = None, 
                 implements: List[str] = None, description: str = None):
        self.name = name
        self.fields = fields or {}
        self.implements = implements or []
        self.description = description
        
    def add_field(self, name: str, type_info: Dict[str, Any]):
        """添加字段到类型"""
        self.fields[name] = type_info
        
    def __str__(self):
        return f"GraphQLType(name={self.name}, fields={len(self.fields)})"


class GraphQLSchema:
    """表示GraphQL Schema的类"""
    
    def __init__(self):
        self.types: Dict[str, GraphQLType] = {}
        self.query_type: Optional[str] = None
        self.mutation_type: Optional[str] = None
        self.subscription_type: Optional[str] = None
        self.interfaces: Dict[str, GraphQLType] = {}
        self.root_fields: Dict[str, Dict[str, Any]] = {}
        self.scalar_types: Set[str] = {'String', 'Int', 'Float', 'Boolean', 'ID'}
        self.enum_types: Set[str] = set()
        self.enum_values: Dict[str, Set[str]] = {}
        
    def add_type(self, type_obj: GraphQLType):
        """添加类型到Schema"""
        self.types[type_obj.name] = type_obj
        
    def add_interface(self, interface_obj: GraphQLType):
        """添加接口到Schema"""
        self.interfaces[interface_obj.name] = interface_obj
        
    def set_query_type(self, name: str):
        """设置查询类型"""
        self.query_type = name
        
    def set_mutation_type(self, name: str):
        """设置变更类型"""
        self.mutation_type = name
        
    def set_subscription_type(self, name: str):
        """设置订阅类型"""
        self.subscription_type = name
        
    def add_root_field(self, name: str, field_info: Dict[str, Any]):
        """添加根字段"""
        self.root_fields[name] = field_info
        
    def add_scalar_type(self, name: str):
        """添加自定义标量类型"""
        self.scalar_types.add(name)
        
    def add_enum_type(self, name: str):
        """添加枚举类型"""
        self.enum_types.add(name)
        
    def add_enum_values(self, type_name: str, values: Set[str]):
        """添加枚举值"""
        self.enum_values[type_name] = values 


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
GraphQL Schema to HttpRunner 转换工具包

该包提供了将GraphQL Schema转换为HttpRunner测试用例的功能。
通过解析GraphQL Schema文件，自动生成适合HttpRunner执行的测试用例。

主要模块：
- models: 数据模型模块，定义了GraphQL Schema相关的数据结构
- parser: 解析器模块，负责解析GraphQL Schema文件
- introspection: 内省查询模块，通过内省查询获取GraphQL Schema
- generator: 生成器模块，负责生成HttpRunner测试用例
- query_generator: 查询语句生成器模块，负责生成GraphQL查询语句列表
- report_generator: 查询语句文件差异报告生成模块，用于在GraphQL Schema发生变更时生成详细的差异比较报告
- main: 主入口模块，提供命令行接口
"""

__version__ = '1.9.3'
__author__ = 'YouMi QA Team'

from .models import GraphQLType, GraphQLSchema
from .parser import GraphQLSchemaParser
from .introspection import fetch_schema_from_introspection, IntrospectionQueryError
from .generator import HttpRunnerTestCaseGenerator
from .query_generator import GraphQLQueryGenerator
from .report_generator import generate_markdown_diff_report, generate_html_diff_report



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
GraphQL 查询语句生成器

该模块提供了从GraphQL Schema生成GraphQL查询语句的功能。
它分析GraphQL Schema中定义的查询、变更和订阅操作，为每个操作生成对应的查询语句。

主要功能：
1. 解析GraphQL Schema，提取所有可用操作
2. 为每个操作生成对应的GraphQL查询语句
3. 支持生成查询、变更和订阅操作的查询语句
4. 支持将生成的查询语句保存到YAML文件
"""

import os
import yaml
from typing import Dict, List, Any, Set

from .models import GraphQLSchema


class GraphQLQueryGenerator:
    """GraphQL查询语句生成器"""
    
    def __init__(self, schema: GraphQLSchema, max_depth: int = 2):
        """
        初始化查询生成器
        
        Args:
            schema: GraphQL Schema对象
            max_depth: 查询嵌套的最大深度
        """
        self.schema = schema
        self.max_depth = max_depth
        self.scalar_types = schema.scalar_types
        self.enum_types = schema.enum_types
        self.generated_types: Set[str] = set()
        
    def generate_queries(self, output_file: str = None, project_name: str = None, query_name: str = None):
        """
        生成所有操作的查询语句并可选择保存到文件
        
        Args:
            output_file: 输出文件路径，如果为None则不保存到文件
            project_name: 项目名称，用于保存时作为分组标识
            query_name: 要生成的查询名称，如果为None则生成所有查询
            
        Returns:
            Dict[str, str]: 操作名称到查询语句的映射
        """
        queries = {}
        
        if query_name:
            # 检查查询名称是否存在
            if query_name not in self.schema.root_fields:
                print(f"错误：查询名称 '{query_name}' 在Schema中不存在")
                return queries
            # 生成指定查询的查询语句
            field_info = self.schema.root_fields[query_name]
            query = self._generate_query(query_name, field_info)
            # 压缩查询语句中的空白，使其单行显示
            query = self._compress_query(query)
            queries[query_name] = query
        else:
            # 生成所有根字段的查询语句
            for field_name, field_info in self.schema.root_fields.items():
                # 跳过名为 _ 的字段，这些字段通常只是占位符
                if field_name == '_':
                    continue
                    
                query = self._generate_query(field_name, field_info)
                # 压缩查询语句中的空白，使其单行显示
                query = self._compress_query(query)
                queries[field_name] = query
        
        # 保存到文件（如果提供了文件路径）
        if output_file:
            output_dir = os.path.dirname(output_file)
            if output_dir and not os.path.exists(output_dir):
                os.makedirs(output_dir, exist_ok=True)            
            
            # 检查文件是否已存在
            file_exists = os.path.exists(output_file)
            
            # 准备写入
            mode = 'a' if file_exists else 'w'
            with open(output_file, mode, encoding='utf-8') as f:
                # 如果是新文件或第一个项目，不需要添加额外换行
                if file_exists:
                    f.write(f"\n{project_name}:\n")
                else:
                    f.write(f"{project_name}:\n")
                # 写入缩进的查询
                for key, value in queries.items():
                    f.write(f"  {key}: '{value}'\n")
                
        return queries
    
    def _compress_query(self, query: str) -> str:
        """
        压缩查询语句，移除多余空白，使其一行显示
        
        Args:
            query: 原始查询语句
            
        Returns:
            str: 压缩后的查询语句
        """
        # 替换所有连续空白为单个空格
        import re
        query = re.sub(r'\s+', ' ', query)
        # 确保花括号前后有空格
        query = re.sub(r'{\s*', '{ ', query)
        query = re.sub(r'\s*}', ' }', query)
        return query.strip()
        
    def _generate_query(self, field_name: str, field_info: Dict[str, Any]) -> str:
        """
        为字段生成查询语句
        
        Args:
            field_name: 字段名称
            field_info: 字段信息
            
        Returns:
            str: 格式化的GraphQL查询语句
        """
        # 确定操作类型
        operation_type = self._determine_operation_type(field_name)
        
        # 构建查询参数
        args_str = ""
        if field_info['args']:
            args_parts = []
            for arg_name, arg_info in field_info['args'].items():
                arg_type = arg_info['type']
                # 生成所有参数，而不仅仅是必填参数
                args_parts.append(f"{arg_name}: ${arg_name}")
                    
            if args_parts:
                args_str = f"({', '.join(args_parts)})"
        
        # 构建变量定义
        variables_def = self._generate_variables_definition(field_name, field_info)
        
        # 构建查询字段
        return_type = field_info['type']
        fields_str = self._generate_fields_for_type(return_type)
        
        # 构建完整查询语句，确保字段之间有空格
        if variables_def:
            query = f"{operation_type} {field_name}{variables_def} {{ {field_name}{args_str} {{ {fields_str} }} }}"
        else:
            query = f"{operation_type} {field_name}{{ {field_name}{args_str} {{ {fields_str} }} }}"
            
        return query
        
    def _determine_operation_type(self, field_name: str) -> str:
        """
        确定字段所属的操作类型（查询、变更或订阅）
        
        Args:
            field_name: 字段名称
            
        Returns:
            str: 操作类型字符串，"query"、"mutation"或"subscription"
        """
        # 检查字段是否定义在Mutation类型中
        if self.schema.mutation_type and field_name in getattr(self.schema.types.get(self.schema.mutation_type), 'fields', {}):
            return "mutation"
        # 检查字段是否定义在Subscription类型中
        elif self.schema.subscription_type and field_name in getattr(self.schema.types.get(self.schema.subscription_type), 'fields', {}):
            return "subscription"
        # 默认为查询操作
        else:
            return "query"
            
    def _generate_variables_definition(self, field_name: str, field_info: Dict[str, Any]) -> str:
        """
        生成变量定义字符串
        
        Args:
            field_name: 字段名称
            field_info: 字段信息
            
        Returns:
            str: 变量定义字符串，形如"($arg1: Type1!, $arg2: Type2)"
        """
        var_defs = []
        
        if field_info['args']:
            for arg_name, arg_info in field_info['args'].items():
                arg_type = arg_info['type']
                # 为所有参数生成变量定义，不仅是必填参数
                var_defs.append(f"${arg_name}: {arg_type}")
                    
        if var_defs:
            return f"({', '.join(var_defs)})"
        return ""
    
    def _generate_fields_for_type(self, type_name: str, depth: int = 0) -> str:
        """
        为类型生成查询字段
        
        Args:
            type_name: 类型名称
            depth: 当前递归深度
            
        Returns:
            str: 格式化的字段字符串
        """
        # 防止无限递归
        if depth > self.max_depth:
            # 不要硬编码返回"id"，而是尝试找到一个合适的字段
            return self._get_safe_field_for_type(type_name)
        
        # 处理列表类型和非空类型
        orig_type_name = type_name
        if type_name.startswith('[') and type_name.endswith(']'):
            inner_type = type_name[1:-1]
            if inner_type.endswith('!'):
                inner_type = inner_type[:-1]
            type_name = inner_type
        
        if type_name.endswith('!'):
            type_name = type_name[:-1]
        
        # 标量和枚举类型不需要子字段
        if type_name in self.scalar_types or type_name in self.enum_types:
            return ""
            
        # 检查类型是否存在
        if type_name not in self.schema.types:
            # 不要硬编码返回"id"，类型不存在时返回空字符串
            return ""
        
        # 防止循环引用
        if type_name in self.generated_types and depth > 0:
            # 不要硬编码返回"id"，而是尝试找到一个合适的字段
            return self._get_safe_field_for_type(type_name)
            
        self.generated_types.add(type_name)
        
        # 获取类型对象
        type_obj = self.schema.types[type_name]
        
        # 构建字段字符串
        fields = []
        
        # 添加id字段（如果存在）
        if 'id' in type_obj.fields:
            fields.append("id")
        
        # 添加其他标量字段和选择的复杂字段
        for field_name, field_info in type_obj.fields.items():
            if field_name == 'id':
                continue
                
            field_type = field_info['type']
            
            # 处理非空标记和列表类型
            if field_type.endswith('!'):
                field_type = field_type[:-1]
                
            if field_type.startswith('[') and field_type.endswith(']'):
                field_type = field_type[1:-1]
                if field_type.endswith('!'):
                    field_type = field_type[:-1]
                
            # 判断是否应该包含该字段
            is_scalar = field_type in self.scalar_types
            is_enum = field_type in self.enum_types
            is_complex_type = field_type in self.schema.types
            has_no_args = not field_info['args']
            
            # 标量类型和枚举类型直接添加
            if is_scalar or is_enum:
                fields.append(field_name)
            # 复杂类型需要递归处理
            elif has_no_args and is_complex_type and depth < self.max_depth:
                sub_fields = self._generate_fields_for_type(field_type, depth + 1)
                if sub_fields:
                    fields.append(f"{field_name} {{ {sub_fields} }}")
        
        # 确保至少有一个字段
        if not fields:
            # 修改逻辑：不假设有id字段，而是选择第一个可用的标量字段
            for field_name, field_info in type_obj.fields.items():
                field_type = field_info['type']
                if field_type.endswith('!'):
                    field_type = field_type[:-1]
                if field_type in self.scalar_types or field_type in self.enum_types:
                    fields.append(field_name)
                    break
        
        # 移除类型，允许在其他分支中使用
        self.generated_types.remove(type_name)
        
        return ", ".join(fields)
    
    def _get_safe_field_for_type(self, type_name: str) -> str:
        """
        为指定类型获取一个安全的字段（优先选择id，如果没有则选择第一个标量字段）
        
        Args:
            type_name: 类型名称
            
        Returns:
            str: 安全的字段名称，如果没有可用字段则返回空字符串
        """
        # 处理类型修饰符
        if type_name.startswith('[') and type_name.endswith(']'):
            inner_type = type_name[1:-1]
            if inner_type.endswith('!'):
                inner_type = inner_type[:-1]
            type_name = inner_type
        
        if type_name.endswith('!'):
            type_name = type_name[:-1]
        
        # 如果类型不存在，返回空字符串
        if type_name not in self.schema.types:
            return ""
        
        type_obj = self.schema.types[type_name]
        
        # 优先返回id字段（如果存在）
        if 'id' in type_obj.fields:
            return "id"
        
        # 如果没有id字段，选择第一个标量或枚举字段
        for field_name, field_info in type_obj.fields.items():
            field_type = field_info['type']
            if field_type.endswith('!'):
                field_type = field_type[:-1]
            if field_type in self.scalar_types or field_type in self.enum_types:
                return field_name
        
        # 如果没有任何标量字段，返回空字符串
        return "" 